# Copyright 2015-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa, Ask Hjorth Larsen, Jens Jørgen Mortensen
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import division
import os
import logging
from contextlib import contextmanager
import numpy as np
from ase.data import chemical_symbols
# import setup_paths
from nomadcore.unit_conversion.unit_conversion import convert_unit as cu
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from gpawparser.tar import Reader
from gpawparser.libxc_names import get_libxc_xc_names
from gpawparser.versions import get_prog_version

@contextmanager
def open_section(p, name):
    gid = p.openSection(name)
    yield gid
    p.closeSection(name, gid)


def c(value, unit=None):
    """ Dummy function for unit conversion"""
    return cu(value, unit)


parser_info = {"name": "parser_gpaw", "version": "1.0"}
path = '../../../../nomad-meta-info/meta_info/nomad_meta_info/' +\
        'gpaw.nomadmetainfo.json'

class GPAWParserWrapper():
    """ A proper class envolop for running this parser using Noamd-FAIRD infra. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        logging.info('GPAW parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("gpaw.nomadmetainfo.json")
        backend = parse_without_class(mainfile, backend)
        return backend

def parse_without_class(filename, backend):
    p = backend
    o = open_section
    r = Reader(filename)
    p.startedParsingSession(filename, parser_info)

    with o(p, 'section_run'):
        p.addValue('program_name', 'GPAW')
        p.addValue('program_version', get_prog_version(r.version))
        if r.Mode == 'pw':
            p.addValue('program_basis_set_type', 'plane waves')
            with o(p, 'section_basis_set_cell_dependent'):
                p.addValue('basis_set_cell_dependent_name',
                           'PW_%.1f_Ry' % (r.PlaneWaveCutoff * 2.0))  # in Ry
                p.addRealValue('basis_set_planewave_cutoff',
                               c(r.PlaneWaveCutoff, 'hartree'))
        elif r.Mode == 'fd':
            with o(p, 'section_basis_set_cell_dependent'):
                h1 = np.linalg.norm(r.UnitCell[0]) / r.dims['ngptsx']
                h2 = np.linalg.norm(r.UnitCell[1]) / r.dims['ngptsy']
                h3 = np.linalg.norm(r.UnitCell[2]) / r.dims['ngptsz']
                h = (h1 + h2 + h3) / 3.0
                p.addValue('basis_set_cell_dependent_name',
                           'GR_%.1f' % (c(h, 'bohr') * 1.0E15))  # in fm
        elif r.Mode == 'lcao':
            p.addValue('program_basis_set_type', 'numeric AOs')
            with o(p, 'section_basis_set_atom_centered'):
                p.addValue('basis_set_atom_centered_short_name', r.BasisSet)
        with o(p, 'section_system') as system_gid:
            p.addArrayValues('simulation_cell', c(r.UnitCell, 'bohr'))
            symbols = np.array([chemical_symbols[z] for z in r.AtomicNumbers])
            p.addArrayValues('atom_labels', symbols)
            p.addArrayValues('atom_positions', c(r.CartesianPositions, 'bohr'))
            if r.Mode == 'pw':
                pbc = np.array((1, 1, 1), bool)
            else:
                pbc = np.array(r.BoundaryConditions, bool)
            p.addArrayValues('configuration_periodic_dimensions', pbc)
        with o(p, 'section_sampling_method'):
            p.addValue('ensemble_type', 'NVE')
        with o(p, 'section_frame_sequence'):
            pass
        with o(p, 'section_method') as method_gid:
            p.addValue('relativity_method', 'pseudo_scalar_relativistic')
            p.addValue('electronic_structure_method', 'DFT')
            p.addValue('scf_threshold_energy_change', c(r.EnergyError,
                                                        'hartree'))
            if 'FixMagneticMoment' in r:
                p.addValue('x_gpaw_fix_magnetic_moment',
                           r.FixMagneticMoment)
                if r.FixMagneticMoment:
                    p.addValue('x_gpaw_fixed_spin_Sz',
                               r.MagneticMoments.sum() / 2.)
            if 'FermiWidth' in r:
                p.addValue('smearing_kind', 'fermi')
                p.addRealValue('smearing_width',
                               c(r.FermiWidth, 'hartree'))
            if 'FixDensity' in r:
                p.addValue('x_gpaw_fix_density', r.FixDensity)
            if 'DensityConvergenceCriterion' in r:
                p.addRealValue('x_gpaw_density_convergence_criterion',
                               r.DensityConvergenceCriterion)
            if 'MixClass' in r:
                p.addValue('x_gpaw_mix_class', r.MixClass)
            if 'MixBeta' in r:
                p.addValue('x_gpaw_mix_beta', r.MixBeta)
            if 'MixWeight' in r:
                p.addValue('x_gpaw_mix_weight', r.MixWeight)
            if 'MixOld' in r:
                p.addValue('x_gpaw_mix_old', r.MixOld)
            if 'MaximumAngularMomentum' in r:
                p.addValue('x_gpaw_maximum_angular_momentum',
                           r.MaximumAngularMomentum)
            if 'SymmetryTimeReversalSwitch' in r:
                p.addValue('x_gpaw_symmetry_time_reversal_switch',
                           r.SymmetryTimeReversalSwitch)
            p.addValue('x_gpaw_xc_functional', r.XCFunctional)
            xc_names = get_libxc_xc_names(r.XCFunctional)
            for name in xc_names.values():
                if name is not None:
                    with o(p, 'section_XC_functionals'):
                        p.addValue('XC_functional_name', name)
        with o(p, 'section_single_configuration_calculation'):
            p.addValue('single_configuration_calculation_to_system_ref',
                       system_gid)
            p.addValue('single_configuration_to_calculation_method_ref',
                       method_gid)
            p.addRealValue('energy_total', c(r.Epot, 'hartree'))
            p.addRealValue('energy_XC', c(r.Exc, 'hartree'))
            p.addRealValue('electronic_kinetic_energy', c(r.Ekin, 'hartree'))
            p.addRealValue('energy_correction_entropy', c(r.S, 'hartree'))
            if 'CartesianForces' in r:
                p.addArrayValues('atom_forces_free',
                                 c(r.CartesianForces, 'bohr/hartree'))
            p.addArrayValues('x_gpaw_magnetic_moments', r.MagneticMoments)
            p.addRealValue('x_gpaw_spin_Sz', r.MagneticMoments.sum() / 2.0)
            #p.addArrayValues('x_gpaw_atomic_density_matrices',
            #                 r.AtomicDensityMatrices)
            #p.addArrayValues('x_gpaw_projections_real', r.Projections.real)
            #p.addArrayValues('x_gpaw_projections_imag', r.Projections.imag)
            with o(p, 'section_eigenvalues'):
                p.addValue('eigenvalues_kind', 'normal')
                p.addArrayValues('eigenvalues_values',
                                 c(r.Eigenvalues, 'hartree'))
                p.addArrayValues('eigenvalues_occupation', r.OccupationNumbers)
                p.addArrayValues('eigenvalues_kpoints', r.IBZKPoints)
    p.finishedParsingSession("ParseSuccess", None)
    return p  # Return the backend

if __name__ == '__main__':
    import sys
    filename = sys.argv[1]
    parse(filename)



# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import numpy as np
from yaml import Loader, YAMLError
from yaml import ScalarNode, SequenceNode, MappingNode, MappingEndEvent
from nomadcore.baseclasses import AbstractBaseParser
from bigdftparser.generic.libxc_codes import LIB_XC_MAPPING
import ase.data
LOGGER = logging.getLogger("nomad")


class BigDFTMainParser(AbstractBaseParser):
    """The main parser class that is called for all run types. Parses the NWChem
    output file.
    """
    def __init__(self, parser_context):
        """
        """
        super(BigDFTMainParser, self).__init__(parser_context)

        # Map keys in the output to funtions that handle the values
        self.key_to_funct_map = {
            "Version Number": lambda x: self.backend.addValue("program_version", x),
            "Atomic structure": self.atomic_structure,
            "Sizes of the simulation domain": self.simulation_domain,
            "Atomic System Properties": self.atomic_system_properties,
            "dft": self.dft,
            "DFT parameters": self.dft_parameters,
            "Ground State Optimization": self.ground_state_optimization,
            "Atomic Forces (Ha/Bohr)": self.atomic_forces,
            "Energy (Hartree)": lambda x: self.backend.addRealValue("energy_total", float(x), unit="hartree"),
        }

    def parse(self, filepath):
        """The output file of a BigDFT run is a YAML document. Here we directly
        parse this document with an existing YAML library, and push its
        contents into the backend. This function will read the document in
        smaller pieces, thus preventing the parser from opening too large files
        directly into memory.
        """
        self.prepare()
        self.print_json_header()
        with open(filepath, "r") as fin:
            try:
                # Open default sections and output default information
                section_run_id = self.backend.openSection("section_run")
                section_system_id = self.backend.openSection("section_system")
                section_method_id = self.backend.openSection("section_method")
                section_scc_id = self.backend.openSection("section_single_configuration_calculation")
                self.backend.addValue("program_name", "BigDFT")
                self.backend.addValue("electronic_structure_method", "DFT")
                self.backend.addValue("program_basis_set_type", "real-space grid")
                self.backend.addValue("single_configuration_calculation_to_system_ref", section_system_id)
                self.backend.addValue("single_configuration_to_calculation_method_ref", section_method_id)

                loader = Loader(fin)
                generator = self.generate_root_nodes(loader)

                # Go through all the keys in the mapping, and call an appropriate
                # function on the value.
                for key, value in generator:

                    function = self.key_to_funct_map.get(key)
                    if function is not None:
                        function(value)

                # Close default sections
                self.backend.closeSection("section_single_configuration_calculation", section_scc_id)
                self.backend.closeSection("section_method", section_method_id)
                self.backend.closeSection("section_system", section_system_id)
                self.backend.closeSection("section_run", section_run_id)
            except YAMLError:
                raise Exception("There was a syntax error in the BigDFT YAML output file.")

        self.print_json_footer()

    def generate_root_nodes(self, loader):
        # Ignore the first two events
        loader.get_event()  # StreamStarEvetn
        loader.get_event()  # DocumentStartEvent
        start_event = loader.get_event()  # MappingStartEvent
        tag = start_event.tag

        # This is the root mapping that contains everything
        node = MappingNode(tag, [],
                start_event.start_mark, None,
                flow_style=start_event.flow_style)

        while not loader.check_event(MappingEndEvent):
            key = loader.construct_scalar(loader.compose_node(node, None))
            value = loader.compose_node(node, key)
            if isinstance(value, MappingNode):
                value = loader.construct_mapping(value, deep=True)
            elif isinstance(value, SequenceNode):
                value = loader.construct_sequence(value, deep=True)
            elif isinstance(value, ScalarNode):
                value = loader.construct_scalar(value)
            yield (key, value)

    #===========================================================================
    # The following functions handle the different sections in the output
    def ground_state_optimization(self, value):
        subspace_optimization = value[0]["Hamiltonian Optimization"]
        subspace_optimization = subspace_optimization[0]["Subspace Optimization"]
        wavefunction_iterations = subspace_optimization["Wavefunctions Iterations"]
        n_iterations = len(wavefunction_iterations)
        self.backend.addValue("number_of_scf_iterations", n_iterations)
        for iteration in wavefunction_iterations:

            scf_id = self.backend.openSection("section_scf_iteration")
            energies = iteration["Energies"]
            # ekin = energies["Ekin"]
            # epot = energies["Epot"]
            # enl = energies["Enl"]
            # eh = energies["EH"]
            # gradient = iteration["gnrm"]
            exc = energies.get("EXC")  # Use get instead, because this value is not always present
            evxc = energies.get("EvXC")  # Use get instead, because this value is not always present
            etotal = iteration["EKS"]
            energy_change = iteration["D"]
            self.backend.addRealValue("energy_total_scf_iteration", etotal, unit="hartree")
            if exc is not None:
                self.backend.addRealValue("energy_XC_scf_iteration", exc, unit="hartree")
            if evxc is not None:
                self.backend.addRealValue("energy_XC_potential_scf_iteration", evxc, unit="hartree")
            self.backend.addRealValue("energy_change_scf_iteration", energy_change, unit="hartree")
            self.backend.closeSection("section_scf_iteration", scf_id)

    def atomic_structure(self, value):
        np_positions = []
        np_labels = []
        positions = value["Positions"]
        for position in positions:
            for key, value in position.items():
                # Not all keys are chemical symbols, e.g. spin, etc.
                if key in ase.data.chemical_symbols:
                    np_positions.append(value)
                    np_labels.append(key)
        np_positions = np.array(np_positions)
        np_labels = np.array(np_labels)
        self.backend.addArrayValues("atom_positions", np_positions, unit="angstrom")
        self.backend.addArrayValues("atom_labels", np_labels)

    def simulation_domain(self, value):
        simulation_cell = np.diag(value["Angstroem"])
        self.backend.addArrayValues("simulation_cell", simulation_cell, unit="angstrom")

    def atomic_system_properties(self, value):
        # Number of atoms
        n_atoms = value["Number of atoms"]
        self.backend.addValue("number_of_atoms", n_atoms)

        # Periodicity
        boundary = value["Boundary Conditions"]
        if boundary == "Free":
            periodic_dimensions = np.array([False, False, False])
        elif boundary == "Periodic":
            periodic_dimensions = np.array([True, True, True])
        elif boundary == "Surface":
            periodic_dimensions = np.array([True, False, True])
        else:
            raise Exception("Unknown boundary condtions.")
        self.backend.addArrayValues("configuration_periodic_dimensions", periodic_dimensions)

    def dft(self, value):
        # Total_charge
        charge = value["qcharge"]
        self.backend.addValue("total_charge", charge)

        # SCF options
        max_iter = value["itermax"]
        self.backend.addValue("scf_max_iteration", max_iter)

        # Spin channels
        n_spin = value["nspin"]
        self.backend.addValue("number_of_spin_channels", n_spin)

    def atomic_forces(self, value):
        forces = []
        for force in value:
            forces.append(*force.values())
        forces = np.array(forces)
        self.backend.addArrayValues("atom_forces", forces, unit="hartree/bohr")

    def dft_parameters(self, value):
        # XC functional
        exchange_settings = value["eXchange Correlation"]
        xc_id = exchange_settings["XC ID"]

        # LibXC codes, see http://bigdft.org/Wiki/index.php?title=XC_codes
        if xc_id < 0:
            xc_parts = []
            xc_id_str = str(xc_id)[1:]
            xc_id_str_len = len(xc_id_str)
            if xc_id_str_len <= 3:
                xc_id_str = "0"*(3-xc_id_str_len)+xc_id_str
                xc_parts.append(xc_id_str)
            elif xc_id_str_len == 6:
                xc_parts.append(xc_id_str[:3])
                xc_parts.append(xc_id_str[3:])

            xc = []
            for part in xc_parts:
                xc1 = LIB_XC_MAPPING.get(part)
                if xc1 is not None:
                    xc.append(xc1)
                else:
                    raise Exception("Unknown LibXC functional number: '{}'".format(part))
        # ABINIT codes, see
        # http://www.tddft.org/programs/octopus/wiki/index.php/Developers_Manual:ABINIT
        # and
        # http://bigdft.org/Wiki/index.php?title=XC_codes
        else:
            mapping = {
                1: ["LDA_XC_TETER93"],
                # 2: ["LDA_C_PZ"],  # Not really sure...
                # 3: Unknown
                # 4: ["LDA_C_WIGNER"],  # Not really sure...
                # 5: ["LDA_C_HL"],  # Not really sure...
                # 6: ["LDA_C_XALPHA"],  # Not really sure...
                # 7: ["LDA_XC_PW"],  # Not really sure...
                # 8: ["LDA_X_PW"],  # Not really sure...
                # 9: ["LDA_X_PW", "LDA_C_RPA"],  # Not really sure...
                # 10: Internal
                11: ["GGA_C_PBE", "GGA_X_PBE"],
                12: ["GGA_X_PBE"],
                # 13: ["GGA_C_PBE","GGA_X_LB"],  # Not really sure...
                # 14: ["GGA_C_PBE","GGA_X_PBE_R"],  # Not really sure...
                15: ["GGA_C_PBE", "GGA_X_RPBE"],
                16: ["GGA_XC_HCTH_93"],
                17: ["GGA_XC_HCTH_120"],
                # 20: Unknown
                # 21: Unknown
                # 22: Unknown
                # 23: ["GGA_X_WC"],  # Not really sure...
                # 24: ["GGA_X_C09X"],  # Not really sure...
                # 25: Internal
                26: ["GGA_XC_HCTH_147"],
                27: ["GGA_XC_HCTH_407"],
                # 28: Internal
                100: ["HF_X"],
            }
            xc = mapping.get(xc_id)

        # Create the XC sections and a summary
        if xc is None:
            raise Exception("Unknown functional number: '{}'".format(xc_id))
        sorted_xc = sorted(xc)
        summary = ""
        n_names = len(sorted_xc)
        for i_name, name in enumerate(sorted_xc):
            weight = 1.0
            xc_id = self.backend.openSection("section_XC_functionals")
            self.backend.addValue("XC_functional_name", name)
            self.backend.addValue("XC_functional_weight", weight)
            self.backend.closeSection("section_XC_functionals", xc_id)
            summary += "{}*{}".format(weight, name)
            if i_name+1 != n_names:
                summary += "_"
        self.backend.addValue("XC_functional", summary)



"""This module should reproduce JSON API v1.0 https://jsonapi.org/format/1.0/"""
# pylint: disable=no-self-argument
from typing import Optional, Union, List
from pydantic import (  # pylint: disable=no-name-in-module
    BaseModel,
    AnyUrl,
    Field,
    root_validator,
)


__all__ = (
    "Meta",
    "Link",
    "JsonApi",
    "ToplevelLinks",
    "ErrorLinks",
    "ErrorSource",
    "BaseResource",
    "RelationshipLinks",
    "Relationship",
    "Relationships",
    "ResourceLinks",
    "Attributes",
    "Resource",
    "Response",
)


class Meta(BaseModel):
    """Non-standard meta-information that can not be represented as an attribute or relationship."""

    class Config:
        extra = "allow"


class Link(BaseModel):
    """A link **MUST** be represented as either: a string containing the link's URL or a link object."""

    href: AnyUrl = Field(..., description="a string containing the link’s URL.")
    meta: Optional[Meta] = Field(
        None,
        description="a meta object containing non-standard meta-information about the link.",
    )


class JsonApi(BaseModel):
    """An object describing the server's implementation"""

    version: str = Field(default="1.0", description="Version of the json API used")
    meta: Optional[Meta] = Field(None, description="Non-standard meta information")


class ToplevelLinks(BaseModel):
    """A set of Links objects, possibly including pagination"""

    self: Optional[Union[AnyUrl, Link]] = Field(None, description="A link to itself")
    related: Optional[Union[AnyUrl, Link]] = Field(
        None, description="A related resource link"
    )

    # Pagination
    first: Optional[AnyUrl] = Field(None, description="The first page of data")
    last: Optional[AnyUrl] = Field(None, description="The last page of data")
    prev: Optional[AnyUrl] = Field(None, description="The previous page of data")
    next: Optional[AnyUrl] = Field(None, description="The next page of data")


class ErrorLinks(BaseModel):
    """A Links object specific to Error objects"""

    about: Optional[Union[AnyUrl, Link]] = Field(
        None,
        description="A link that leads to further details about this particular occurrence of the problem.",
    )


class ErrorSource(BaseModel):
    """an object containing references to the source of the error"""

    pointer: Optional[str] = Field(
        None,
        description="a JSON Pointer [RFC6901] to the associated entity in the request document "
        '[e.g. "/data" for a primary data object, or "/data/attributes/title" for a specific attribute].',
    )
    parameter: Optional[str] = Field(
        None,
        description="a string indicating which URI query parameter caused the error.",
    )


class Error(BaseModel):
    """An error response"""

    id: Optional[str] = Field(
        None,
        description="A unique identifier for this particular occurrence of the problem.",
    )
    links: Optional[ErrorLinks] = Field(
        None, description="A links object storing about"
    )
    status: Optional[str] = Field(
        None,
        description="the HTTP status code applicable to this problem, expressed as a string value.",
    )
    code: Optional[str] = Field(
        None,
        description="an application-specific error code, expressed as a string value.",
    )
    title: Optional[str] = Field(
        None,
        description="A short, human-readable summary of the problem. "
        "It **SHOULD NOT** change from occurrence to occurrence of the problem, except for purposes of localization.",
    )
    detail: Optional[str] = Field(
        None,
        description="A human-readable explanation specific to this occurrence of the problem.",
    )
    source: Optional[ErrorSource] = Field(
        None, description="An object containing references to the source of the error"
    )
    meta: Optional[Meta] = Field(
        None,
        description="a meta object containing non-standard meta-information about the error.",
    )

    def __hash__(self):
        return hash(self.json())


class BaseResource(BaseModel):
    """Minimum requirements to represent a Resource"""

    id: str = Field(..., description="Resource ID")
    type: str = Field(..., description="Resource type")


class RelationshipLinks(BaseModel):
    """A resource object **MAY** contain references to other resource objects (\"relationships\").
    Relationships may be to-one or to-many. Relationships can be specified by including a member in a resource's links object."""

    self: Optional[Union[AnyUrl, Link]] = Field(None, description="A link to itself")
    related: Optional[Union[AnyUrl, Link]] = Field(
        None, description="A related resource link"
    )

    @root_validator(pre=True)
    def either_self_or_related_must_be_specified(cls, values):
        for value in values.values():
            if value is not None:
                break
        else:
            raise ValueError(
                "Either 'self' or 'related' MUST be specified for RelationshipLinks"
            )
        return values


class Relationship(BaseModel):
    """Representation references from the resource object in which it’s defined to other resource objects."""

    links: Optional[RelationshipLinks] = Field(
        None,
        description="a links object containing at least one of the following: self, related",
    )
    data: Optional[Union[BaseResource, List[BaseResource]]] = Field(
        None, description="Resource linkage", uniqueItems=True
    )
    meta: Optional[Meta] = Field(
        None,
        description="a meta object that contains non-standard meta-information about the relationship.",
    )

    @root_validator(pre=True)
    def at_least_one_relationship_key_must_be_set(cls, values):
        for value in values.values():
            if value is not None:
                break
        else:
            raise ValueError(
                "Either 'links', 'data', or 'meta' MUST be specified for Relationship"
            )
        return values


class Relationships(BaseModel):
    """
    Members of the relationships object (\"relationships\") represent references from the resource object in which it's defined to other resource objects.
    Keys MUST NOT be:
        type
        id
    """

    @root_validator(pre=True)
    def check_illegal_relationships_fields(cls, values):
        illegal_fields = ("id", "type")
        for field in illegal_fields:
            if field in values:
                raise ValueError(
                    f"{illegal_fields} MUST NOT be fields under Relationships"
                )
        return values


class ResourceLinks(BaseModel):
    """A Resource Links object"""

    self: Optional[Union[AnyUrl, Link]] = Field(
        None,
        description="A link that identifies the resource represented by the resource object.",
    )


class Attributes(BaseModel):
    """
    Members of the attributes object ("attributes\") represent information about the resource object in which it's defined.
    The keys for Attributes MUST NOT be:
        relationships
        links
        id
        type
    """

    class Config:
        extra = "allow"

    @root_validator(pre=True)
    def check_illegal_attributes_fields(cls, values):
        illegal_fields = ("relationships", "links", "id", "type")
        for field in illegal_fields:
            if field in values:
                raise ValueError(
                    f"{illegal_fields} MUST NOT be fields under Attributes"
                )
        return values


class Resource(BaseResource):
    """Resource objects appear in a JSON:API document to represent resources."""

    links: Optional[ResourceLinks] = Field(
        None, description="a links object containing links related to the resource."
    )
    meta: Optional[Meta] = Field(
        None,
        description="a meta object containing non-standard meta-information about a resource that can not be represented as an attribute or relationship.",
    )
    attributes: Optional[Attributes] = Field(
        None,
        description="an attributes object representing some of the resource’s data.",
    )
    relationships: Optional[Relationships] = Field(
        None,
        description="a relationships object describing relationships between the resource and other JSON:API resources.",
    )


class Response(BaseModel):
    """A top-level response"""

    data: Optional[Union[None, Resource, List[Resource]]] = Field(
        None, description="Outputted Data", uniqueItems=True
    )
    meta: Optional[Meta] = Field(
        None,
        description="A meta object containing non-standard information related to the Success",
    )
    errors: Optional[List[Error]] = Field(
        None, description="A list of unique errors", uniqueItems=True
    )
    included: Optional[List[Resource]] = Field(
        None, description="A list of unique included resources", uniqueItems=True
    )
    links: Optional[ToplevelLinks] = Field(
        None, description="Links associated with the primary data or errors"
    )
    jsonapi: Optional[JsonApi] = Field(
        None, description="Information about the JSON API used"
    )

    @root_validator(pre=True)
    def either_data_meta_or_errors_must_be_set(cls, values):
        required_fields = ("data", "meta", "errors")
        if not any(values.get(field) for field in required_fields):
            raise ValueError(
                f"At least one of {required_fields} MUST be specified in the top-level response"
            )
        return values



# Copyright 2015-2018 Ask Hjorth Larsen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# coding = utf-8

"""This module is written by

  Ask Hjorth Larsen <asklarsen@gmail.com>
  Carlos de Armas
"""

import os
import re
from subprocess import Popen, PIPE

import numpy as np

from ase import Atoms
from ase.calculators.calculator import FileIOCalculator
from ase.data import atomic_numbers
from ase.io import read
from ase.io.xsf import read_xsf
from ase.units import Bohr, Angstrom, Hartree, eV, Debye


def read_eigenvalues_file(fd):
    unit = None

    for line in fd:
        m = re.match('Eigenvalues\s*\[(.+?)\]', line)
        if m is not None:
            unit = m.group(1)
            break
    line = next(fd)
    assert line.strip().startswith('#st'), line

    kpts = []
    eigs = []
    occs = []

    for line in fd:
        m = re.match(r'#k.*?\(\s*(.+?),\s*(.+?),\s*(.+?)\)', line)
        if m:
            k = m.group(1, 2, 3)
            kpts.append(np.array(k, float))
            eigs.append({})
            occs.append({})
        else:
            m = re.match(r'\s*\d+\s*(\S+)\s*(\S+)\s*(\S+)', line)
            assert m is not None
            spin, eig, occ = m.group(1, 2, 3)
            eigs[-1].setdefault(spin, []).append(float(eig))
            occs[-1].setdefault(spin, []).append(float(occ))

    nkpts = len(kpts)
    nspins = len(eigs[0])
    nbands = len(eigs[0][spin])

    kptsarr = np.array(kpts, float)
    eigsarr = np.empty((nkpts, nspins, nbands))
    occsarr = np.empty((nkpts, nspins, nbands))

    arrs = [eigsarr, occsarr]

    for arr in arrs:
        arr.fill(np.nan)

    for k in range(nkpts):
        for arr, lst in [(eigsarr, eigs), (occsarr, occs)]:
            arr[k, :, :] = [lst[k][spin] for spin
                            in (['--'] if nspins == 1 else ['up', 'dn'])]

    for arr in arrs:
        assert not np.isnan(arr).any()

    eigsarr *= {'H': Hartree, 'eV': eV}[unit]
    return kptsarr, eigsarr, occsarr


class OctopusIOError(IOError):
    pass  # Cannot find output files


def unpad(pbc, arr):
    # Return non-padded array from padded array.
    # This means removing the last element along all periodic directions.
    if pbc[0]:
        assert np.all(arr[0, :, :] == arr[-1, :, :])
        arr = arr[0:-1, :, :]
    if pbc[1]:
        assert np.all(arr[:, 0, :] == arr[:, -1, :])
        arr = arr[:, 0:-1, :]
    if pbc[2]:
        assert np.all(arr[:, :, 0] == arr[:, :, -1])
        arr = arr[:, :, 0:-1]
    return np.ascontiguousarray(arr)


# Parse value as written in input file *or* something that one would be
# passing to the ASE interface, i.e., this might already be a boolean
def octbool2bool(value):
    value = value.lower()
    if isinstance(value, int):
        return bool(value)
    if value in ['true', 't', 'yes', '1']:
        return True
    elif value in ['no', 'f', 'false', '0']:
        return False
    else:
        raise ValueError('Failed to interpret "%s" as a boolean.' % value)


def list2block(name, rows):
    """Construct 'block' of Octopus input.

    convert a list of rows to a string with the format x | x | ....
    for the octopus input file"""
    lines = []
    lines.append('%' + name)
    for row in rows:
        lines.append(' ' + ' | '.join(str(obj) for obj in row))
    lines.append('%')
    return lines


def normalize_keywords(kwargs):
    """Reduce keywords to unambiguous form (lowercase)."""
    newkwargs = {}
    for arg, value in kwargs.items():
        lkey = arg.lower()
        newkwargs[lkey] = value
    return newkwargs


def get_octopus_keywords():
    """Get dict mapping all normalized keywords to pretty keywords."""
    proc = Popen(['oct-help', '--search', ''], stdout=PIPE)
    keywords = proc.stdout.read().decode().split()
    return normalize_keywords(dict(zip(keywords, keywords)))


def input_line_iter(lines):
    """Convenient iterator for parsing input files 'cleanly'.

    Discards comments etc."""
    for line in lines:
        line = line.split('#')[0].strip()
        if not line or line.isspace():
            continue
        line = line.strip()
        yield line


def block2list(namespace, lines, header=None):
    """Parse lines of block and return list of lists of strings."""
    lines = iter(lines)
    block = []
    if header is None:
        header = next(lines)
    assert header.startswith('%'), header
    name = header[1:].strip().lower()
    for line in lines:
        if line.startswith('%'):  # Could also say line == '%' most likely.
            break
        tokens = [namespace.evaluate(token)
                  for token in line.strip().split('|')]
        # XXX will fail for string literals containing '|'
        block.append(tokens)
    return name, block


class OctNamespace:
    def __init__(self):
        self.names = {}
        self.consts = {'pi': np.pi,
                       'angstrom': 1. / Bohr,
                       'ev': 1. / Hartree,
                       'yes': True,
                       'no': False,
                       't': True,
                       'f': False,
                       'i': 1j,  # This will probably cause trouble
                       'true': True,
                       'false': False}

    def evaluate(self, value):
        orig = value
        value = value.strip()

        for char in '"', "'":  # String literal
            if value.startswith(char):
                assert value.endswith(char)
                return value

        value = value.lower()

        if value in self.consts:  # boolean or other constant
            return self.consts[value]

        if value in self.names:  # existing variable
            return self.names[value]

        try:  # literal integer
            v = int(value)
        except ValueError:
            pass
        else:
            if v == float(v):
                return v

        try:  # literal float
            return float(value)
        except ValueError:
            pass

        if ('*' in value or '/' in value
            and not any(char in value for char in '()+')):
            floatvalue = 1.0
            op = '*'
            for token in re.split(r'([\*/])', value):
                if token in '*/':
                    op = token
                    continue

                v = self.evaluate(token)

                try:
                    v = float(v)
                except TypeError:
                    try:
                        v = complex(v)
                    except ValueError:
                        break
                except ValueError:
                    break  # Cannot evaluate expression
                else:
                    if op == '*':
                        floatvalue *= v
                    else:
                        assert op == '/', op
                        floatvalue /= v
            else:  # Loop completed successfully
                return floatvalue
        return value  # unknown name, or complex arithmetic expression

    def add(self, name, value):
        value = self.evaluate(value)
        self.names[name.lower().strip()] = value


def parse_input_file(fd):
    namespace = OctNamespace()
    lines = input_line_iter(fd)
    blocks = {}
    while True:
        try:
            line = next(lines)
        except StopIteration:
            break
        else:
            if line.startswith('%'):
                name, value = block2list(namespace, lines, header=line)
                blocks[name] = value
            else:
                tokens = line.split('=', 1)
                assert len(tokens) == 2, tokens
                name, value = tokens
                namespace.add(name, value)

    namespace.names.update(blocks)
    return namespace.names


def read_static_info_kpoints(fd):
    for line in fd:
        if line.startswith('List of k-points'):
            break

    tokens = next(fd).split()
    assert tokens == ['ik', 'k_x', 'k_y', 'k_z', 'Weight']
    bar = next(fd)
    assert bar.startswith('---')

    kpts = []
    weights = []

    for line in fd:
        # Format:        index   kx      ky      kz     weight
        m = re.match(r'\s*\d+\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)', line)
        if m is None:
            break
        kxyz = m.group(1, 2, 3)
        weight = m.group(4)
        kpts.append(kxyz)
        weights.append(weight)

    ibz_k_points = np.array(kpts, float)
    k_point_weights = np.array(weights, float)
    return dict(ibz_k_points=ibz_k_points, k_point_weights=k_point_weights)


def read_static_info_eigenvalues_efermi(fd, energy_unit):
    '''
    Parse eigenvalues and Fermi Energy from `static/info`
    fd: file descriptor
    energy_unit: string
    Returns: dictionary
    '''

    # we had to allow this function to handle both
    # eigenvalues and Fermi energies because the later
    # comes in the output file immediately after the last
    # eigenvalue, hence the "fermi line" is consumed by the
    # line checkers that look for eigenvalues.

    values_sknx = {}

    eFermi = None
    nbands = 0
    for line in fd:
        line = line.strip()
        if line.startswith('#'):
            continue
        if line.startswith('Fermi'):  # tmk
            # print(line, '##') #  OK!
            tokens = line.split()
            unit = {'eV': eV, 'H': Hartree}[tokens[-1]]
            eFermi = float(tokens[-2]) * unit
        if not line[:1].isdigit():
            break

        tokens = line.split()
        nbands = max(nbands, int(tokens[0]))
        energy = float(tokens[2]) * energy_unit
        occupation = float(tokens[3])
        values_sknx.setdefault(tokens[1], []).append((energy, occupation))

    nspins = len(values_sknx)
    if nspins == 1:
        val = [values_sknx['--']]
    else:
        val = [values_sknx['up'], values_sknx['dn']]
    val = np.array(val).astype(float)
    nkpts, remainder = divmod(len(val[0]), nbands)
    assert remainder == 0

    eps_skn = val[:, :, 0].reshape(nspins, nkpts, nbands)
    occ_skn = val[:, :, 1].reshape(nspins, nkpts, nbands)
    eps_skn = eps_skn.transpose(1, 0, 2).copy()
    occ_skn = occ_skn.transpose(1, 0, 2).copy()
    assert eps_skn.flags.contiguous
    # print('save to dictionary: ', nspins, nkpts, nbands, eps_skn, occ_skn, eFermi)
    info_dict = dict(nspins=nspins,
                     nkpts=nkpts,
                     nbands=nbands,
                     eigenvalues=eps_skn,
                     occupations=occ_skn)
    if eFermi is not None:
        # not all Octopus' output might report the Fermi energy
        info_dict['efermi'] = eFermi
    return info_dict


def read_static_info_energy(fd, energy_unit):
    def get(name):
        for line in fd:
            if line.strip().startswith(name):
                return float(line.split('=')[-1].strip()) * energy_unit
    return dict(energy=get('Total'), free_energy=get('Free'))


def read_static_info(fd):
    # fd: file descriptor
    results = {}
    def get_energy_unit(line):  # Convert "title [unit]": ---> unit
        return {'[eV]': eV, '[H]': Hartree}[line.split()[1].rstrip(':')]

    for line in fd:
        if line.strip('*').strip().startswith('Brillouin zone'):
            results.update(read_static_info_kpoints(fd))
        elif line.startswith('Eigenvalues ['):
            unit = get_energy_unit(line)
            results.update(read_static_info_eigenvalues_efermi(fd, unit))
        elif line.startswith('Energy ['):
            unit = get_energy_unit(line)
            results.update(read_static_info_energy(fd, unit))
        elif line.startswith('Total Magnetic Moment'):
            if 0:
                line = next(fd)
                values = line.split()
                results['magmom'] = float(values[-1])

                line = next(fd)
                assert line.startswith('Local Magnetic Moments')
                line = next(fd)
                assert line.split() == ['Ion', 'mz']
                # Reading  Local Magnetic Moments
                mag_moment = []
                for line in fd:
                    if line == '\n':
                        break  # there is no more thing to search for
                    line = line.replace('\n', ' ')
                    values = line.split()
                    mag_moment.append(float(values[-1]))

                results['magmoms'] = np.array(mag_moment)
        elif line.startswith('Dipole'):
            assert line.split()[-1] == '[Debye]'
            dipole = [float(next(fd).split()[-1]) for i in range(3)]
            results['dipole'] = np.array(dipole) * Debye
        elif line.startswith('Forces'):
            forceunitspec = line.split()[-1]
            forceunit = {'[eV/A]': eV / Angstrom,
                         '[H/b]': Hartree / Bohr}[forceunitspec]
            forces = []
            line = next(fd)
            assert line.strip().startswith('Ion')
            for line in fd:
                if line.strip().startswith('---'):
                    break
                tokens = line.split()[-3:]
                forces.append([float(f) for f in tokens])
            results['forces'] = np.array(forces) * forceunit

    if 'ibz_k_points' not in results:
        results['ibz_k_points'] = np.zeros((1, 3))
        results['k_point_weights'] = np.ones(1)
    if 'efermi' not in results:
        # Find HOMO level.  Note: This could be a very bad
        # implementation with fractional occupations if the Fermi
        # level was not found otherwise.
        all_energies = results['eigenvalues'].ravel()
        all_occupations = results['occupations'].ravel()
        args = np.argsort(all_energies)
        for arg in args[::-1]:
            if all_occupations[arg] > 0.1:
                break
        eFermi = all_energies[arg]
        results['efermi'] = eFermi
        # print('eFermi estimate', eFermi) # this produces bad estimate
        # ----
    return results


class Octopus(FileIOCalculator):
    def __init__(self, restart):
        self.kwargs = {}

        FileIOCalculator.__init__(self, restart=restart, label=restart)
        # The above call triggers set() so we can update self.kwargs.

    def set_label(self, label):
        # Octopus does not support arbitrary namings of all the output files.
        # But we can decide that we always dump everything in a directory.
        if not label.endswith('/'):
            label += '/'
        FileIOCalculator.set_label(self, label)

    def set(self, **kwargs):
        """Set octopus input file parameters."""
        kwargs = normalize_keywords(kwargs)

        changes = FileIOCalculator.set(self, **kwargs)
        if changes:
            self.results.clear()
        self.kwargs.update(kwargs)
        # XXX should use 'Parameters' but don't know how

    def get_xc_functional(self):
        """Return the XC-functional identifier.
            'LDA', 'PBE', ..."""
        return self.kwargs.get('xcfunctional', 'LDA')

    def get_fermi_level(self):
        return self.results['efermi']

    def get_fermi_energy(self):
        return self.results['efermi']

    def get_dipole_moment(self, atoms=None):
        if 'dipole' not in self.results:
            msg = ('Dipole moment not calculated.\n'
                   'You may wish to use SCFCalculateDipole=True')
            raise OctopusIOError(msg)
        return self.results['dipole']

    def _read_array(self, fname, outputkeyword=None):
        path = self._getpath('static/%s' % fname)
        if not os.path.exists(path):
            msg = 'Path not found: %s' % path
            if outputkeyword is not None:
                msg += ('\nIt appears that the %s has not been saved.\n'
                        'Be sure to specify Output=\'%s\' in the input.'
                        % (outputkeyword, outputkeyword))
            raise OctopusIOError(msg)
        # If this causes an error now that the file exists, things are
        # messed up.  Then it is better that the error propagates as normal
        return read_xsf(path, read_data=True)

    def read_vn(self, basefname, keywordname):
        static_dir = self._getpath('static')
        assert os.path.isdir(static_dir)

        if self.get_spin_polarized():
            spin1, _atoms = self._read_array('%s-sp1.xsf' % basefname,
                                             keywordname)
            spin2, _atoms = self._read_array('%s-sp2.xsf' % basefname,
                                             keywordname)
            array = np.array([spin1, spin2])  # shape 2, nx, ny, nz
        else:
            array, _atoms = self._read_array('%s.xsf' % basefname, keywordname)
            array = array[None]  # shape 1, nx, ny, nx
        assert len(array.shape) == 4
        return array

    def _unpad_periodic(self, array):
        return unpad(self.get_atoms().pbc, array)

    def _pad_unperiodic(self, array):
        pbc = self.get_atoms().pbc
        orig_shape = array.shape
        newshape = [orig_shape[c] + (0 if pbc[c] else 1) for c in range(3)]
        out = np.zeros(newshape, dtype=array.dtype)
        nx, ny, nz = orig_shape
        out[:nx, :ny, :nz] = array
        return out

    def _pad_correctly(self, array, pad):
        array = self._unpad_periodic(array)
        if pad:
            array = self._pad_unperiodic(array)
        return array

    def get_pseudo_density(self, spin=None, pad=True):
        """Return pseudo-density array.

        If *spin* is not given, then the total density is returned.
        Otherwise, the spin up or down density is returned (spin=0 or
        1)."""
        if 'density_sg' not in self.results:
            self.results['density_sg'] = self.read_vn('density', 'density')
        density_sg = self.results['density_sg']
        if spin is None:
            density_g = density_sg.sum(axis=0)
        else:
            assert spin == 0 or (spin == 1 and len(density_sg) == 2)
            density_g = density_sg[spin]
        return self._pad_correctly(density_g, pad)

    def get_effective_potential(self, spin=0, pad=True):
        if spin is None:  # Annoying case because it works as an index!
            raise ValueError('spin=None')
        if 'potential_sg' not in self.results:
            self.results['potential_sg'] = self.read_vn('vks', 'potential')
        array = self.results['potential_sg'][spin]
        return self._pad_correctly(array, pad)

    def get_pseudo_wave_function(self, band=0, kpt=0, spin=0, broadcast=True,
                                 pad=True):
        """Return pseudo-wave-function array."""
        assert band < self.get_number_of_bands()

        ibz_k_pts = self.get_ibz_k_points()

        forcecomplex = self.kwargs.get('forcecomplex')
        if forcecomplex is not None:
            forcecomplex = octbool2bool(forcecomplex)
        if len(ibz_k_pts) > 1 or ibz_k_pts.any() or forcecomplex:
            dtype = complex
        else:
            dtype = float  # Might there be more issues that determine dtype?

        if self.get_spin_polarized():
            kpt_index = 2 * kpt + spin  # XXX this is *probably* correct
        else:
            kpt_index = kpt

        # The ASE convention is that kpts and bands start from 0,
        # whereas in Octopus they start from 1.  So always add 1
        # when looking for filenames.
        kpt_index += 1
        band_index = band + 1

        tokens = ['wf']
        if len(ibz_k_pts) > 1 or self.get_spin_polarized():
            tokens.append('-k%03d' % kpt_index)
        tokens.append('-st%04d' % band_index)
        name = ''.join(tokens)

        if dtype == float:
            array, _atoms = self._read_array('%s.xsf' % name, 'wfs')
        else:
            array_real, _atoms = self._read_array('%s.real.xsf' % name, 'wfs')
            array_imag, _atoms = self._read_array('%s.imag.xsf' % name, 'wfs')
            array = array_real + 1j * array_imag

        return self._pad_correctly(array, pad)

    def get_number_of_spins(self):
        """Return the number of spins in the calculation.
           Spin-paired calculations: 1, spin-polarized calculation: 2."""
        return 2 if self.get_spin_polarized() else 1

    def get_spin_polarized(self):
        """Is it a spin-polarized calculation?"""

        sc = self.kwargs.get('spincomponents')
        if sc is None or sc == 'unpolarized' or sc == 'spinors':
            return False
        elif sc == 'spin_polarized' or sc == 'polarized':
            return True
        # else:
        #    raise NotImplementedError('SpinComponents keyword %s' % sc)

    def get_ibz_k_points(self):
        """Return k-points in the irreducible part of the Brillouin zone.
        The coordinates are relative to reciprocal latice vectors."""
        return self.results['ibz_k_points']

    def get_k_point_weights(self):
        return self.results['k_point_weights']

    def get_number_of_bands(self):
        return self.results['nbands']

    def get_magnetic_moments(self, atoms=None):
        if self.results['nspins'] == 1:
            return np.zeros(len(self.atoms))
        return self.results['magmoms'].copy()

    def get_magnetic_moment(self, atoms=None):
        if self.results['nspins'] == 1:
            return 0.0
        return self.results['magmom']

    def get_occupation_numbers(self, kpt=0, spin=0):
        return self.results['occupations'][kpt, spin].copy()

    def get_eigenvalues(self, kpt=0, spin=0):
        return self.results['eigenvalues'][kpt, spin].copy()

    def _getpath(self, path, check=False):
        path = os.path.join(self.directory, path)
        if check:
            if not os.path.exists(path):
                raise OctopusIOError('No such file or directory: %s' % path)
        return path

    def read_results(self):
        """Read octopus output files and extract data."""
        fd = open(self._getpath('static/info', check=True))
        self.results.update(read_static_info(fd))

        # If the eigenvalues file exists, we get the eigs/occs from that one.
        # This probably means someone ran Octopus in 'unocc' mode to
        # get eigenvalues (e.g. for band structures), and the values in
        # static/info will be the old (selfconsistent) ones.
        try:
            eigpath = self._getpath('static/eigenvalues', check=True)
        except OctopusIOError:
            pass
        else:
            with open(eigpath) as fd:
                kpts, eigs, occs = read_eigenvalues_file(fd)
                kpt_weights = np.ones(len(kpts))  # XXX ?  Or 1 / len(kpts) ?
            self.results.update(eigenvalues=eigs, occupations=occs,
                                ibz_k_points=kpts,
                                k_point_weights=kpt_weights)

    def read(self, label):
        # XXX label of restart file may not be the same as actual label!
        # This makes things rather tricky.  We first set the label to
        # that of the restart file and arbitrarily expect the remaining code
        # to rectify any consequent inconsistencies.
        self.set_label(label)

        FileIOCalculator.read(self, label)
        inp_path = self._getpath('inp')
        fd = open(inp_path)
        kwargs = parse_input_file(fd)

        # self.atoms, kwargs = kwargs2atoms(kwargs)
        self.kwargs.update(kwargs)

        fd.close()
        self.read_results()



# Copyright 2017-2018 Berk Onat, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from nomadcore.local_meta_info import InfoKindEl
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from .AMBERDictionary import get_unitDict, get_nameListDict, get_fileListDict
import nomadcore.metainfo_storage.MetaInfoStorage as mStore
import logging
import re

from .metainfo import m_env as metainfo_env

PARSERNAME = "AMBER"
PROGRAMNAME = "amber"
PARSERVERSION = "0.0.3"
PARSERMETANAME = PARSERNAME.lower()
PARSERTAG = 'x_' + PARSERMETANAME

PARSER_INFO_DEFAULT = {
    'name': 'amber-parser',
    'version': '0.0.1'
}

LOGGER = logging.getLogger("nomad.AMBERParser")


def get_metaInfo(self):
    metaInfoEnv = metainfo_env.legacy_info_env
    metaInfoEnv = set_section_metaInfoEnv(metaInfoEnv, 'section', ['input_output_files'], 'type_section', False, ["section_run"])
    metaInfoEnv = setDict_metaInfoEnv(metaInfoEnv, self.fileDict)
    metaInfoEnv = setDict_metaInfoEnv(metaInfoEnv, self.cntrlDict)
    metaInfoEnv = setDict_metaInfoEnv(metaInfoEnv, self.ewaldDict)
    metaInfoEnv = setDict_metaInfoEnv(metaInfoEnv, self.qmmmDict)
    metaInfoEnv = setDict_metaInfoEnv(metaInfoEnv, self.parmDict)
    metaInfoEnv = setDict_metaInfoEnv(metaInfoEnv, self.mddataDict)
    metaInfoEnv = setDict_metaInfoEnv(metaInfoEnv, self.extraDict)
    return metaInfoEnv


def set_section_metaInfoEnv(infoEnv, metaNameTag, newList, listTypStr, repeatingSection, supraNames):
    """Modifies meta info data.

    Args:
        metaInfoEnv: meta info environment json type data.

    Returns:
        metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
    """
    for newName in newList:
        newName = newName.lower().replace(" ", "").replace("-", "")
        if 'x_amber_%s_%s' % (metaNameTag, newName) not in infoEnv.infoKinds:
            infoEnv.addInfoKindEl(InfoKindEl(
                description='auto generated section meta info data',
                name='x_amber_%s_%s' % (metaNameTag, newName),
                kindStr=listTypStr,
                repeats=repeatingSection,
                superNames=supraNames))

    return infoEnv


def setDict_metaInfoEnv(infoEnv, nameDict):
    """Modifies meta info data.

    Args:
        metaInfoEnv: meta info environment json type data.
        nameDict: dictionary for name info and data.

    Returns:
        metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
    """
    for keyName in nameDict.keys():
        if '%s' % (keyName) not in infoEnv.infoKinds:
            infoEnv.addInfoKindEl(InfoKindEl(
                name='%s' % (keyName),
                description='auto generated meta info data',
                dtypeStr=nameDict[keyName].metaInfoType,
                shape=[],
                superNames=nameDict[keyName].activeSections))

    return infoEnv


def set_metaInfoEnv(infoEnv, metaNameTag, newList, listTypStr, supraNames):
    """Modifies meta info data.

    Args:
        metaInfoEnv: meta info environment json type data.

    Returns:
        metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
    """
    for newName in newList:
        newName = newName.lower().replace(" ", "").replace("-", "")
        if 'x_amber_%s_%s' % (metaNameTag, newName) not in infoEnv.infoKinds:
            infoEnv.addInfoKindEl(InfoKindEl(
                name='x_amber_%s_%s' % (metaNameTag, newName),
                description='auto generated meta info data',
                dtypeStr=listTypStr,
                shape=[],
                superNames=supraNames))

    return infoEnv


class AMBERParserBase(object):
    """Base class for Amber parsers"""
    def __init__(
            self, cachingLevelForMetaName=None, coverageIgnoreList=None, re_program_name=None):
        self.metaStorage = mStore.Container('section_run')
        self.metaStorageRestrict = mStore.Container('section_restricted_uri')
        exclude_dict = {
            'section_run': [
                'section_processor_info',
                'section_processor_log',
                'section_springer_material',
                'section_repository_info'
            ]
        }
        jsonmetadata = mStore.JsonMetaInfo(metainfo_env.legacy_info_env)
        self.metaStorage.build(jsonmetadata, 'section_run', exclude_dict)
        self.metaStorageRestrict.build(jsonmetadata, 'section_restricted_uri', exclude_dict)
        self.re_program_name = re_program_name
        self.unitDict = get_unitDict('si')
        self.fileDict = get_fileListDict()
        self.cntrlDict = get_nameListDict('cntrl')
        self.ewaldDict = get_nameListDict('ewald')
        self.qmmmDict = get_nameListDict('qmmm')
        self.wtDict = get_nameListDict('wt')
        self.parmDict = get_nameListDict('parm')
        self.mddataDict = get_nameListDict('mddata')
        self.extraDict = get_nameListDict('extra')
        self.parserInfo = PARSER_INFO_DEFAULT.copy()
        self.metaInfoEnv = get_metaInfo(self)
        self.coverageIgnoreList = [
            # ignore empty lines
            r"\s*",
            # table separators
            # r"^\s*[=%-]+\s*$",
            # r"^\s*%\s*%\s*$",
        ]
        self.coverageIgnore = None

    def parse(self):
        self.coverageIgnore = re.compile(r"^(?:" + r"|".join(self.coverageIgnoreList) + r")$")
        mainFunction(
            mainFileDescription=self.mainFileDescription(),
            metaInfoEnv=self.metaInfoEnv,
            parserInfo=self.parserInfo,
            #  cachingLevelForMetaName=self.cachingLevelForMetaName,
            superContext=self)

    def adHoc_amber_program_name(self, parser):
        if self.re_program_name is not None:
            if not self.re_program_name.match(
                    parser.lastMatch['x_amber_program_name']):
                raise Exception(
                    "mainFile program name was: %s, unsuited for %s" % (
                        parser.lastMatch['x_amber_program_name'],
                        type(self).__name__))

    def mainFileDescription(self):
        # assemble matchers and submatchers
        return SM(
            name='Root',
            startReStr="",
            forwardMatch=True,
            weak=True,
            subMatchers=self.build_subMatchers())

    def build_subMatchers(self):
        return []



# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#
# Main author and maintainer: Ask Hjorth Larsen <asklarsen@gmail.com>

from __future__ import print_function
import os

from main import main

testdir = '../../test/examples'
dirnames = ['H2O',
            'H2O-relax',
            'Al-slab',
            'Al-uc',
            'Fe',
            'MgO']


for dirname in dirnames:
    fname = os.path.join(testdir, dirname, 'out')
    with open('out.pycheck.%s.txt' % dirname, 'w') as outfd:
        main(mainFile=fname, outF=outfd)



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import general_experimental

m_package = Package(
    name='skeleton_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='skeleton.nomadmetainfo.json'))


class section_experiment(general_experimental.section_experiment):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_experiment'))

    experiment_location = Quantity(
        type=str,
        shape=[],
        description='''
        Contains information relating to an archive.
        ''',
        a_legacy=LegacyDefinition(name='experiment_location'))


m_package.__init_metainfo__()



# Copyright 2017-2018 Lorenzo Pardini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from builtins import object
# import setup_paths
import numpy as np
from nomadcore.simple_parser import SimpleMatcher as SM, mainFunction
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.caching_backend import CachingLevel
from nomadcore.unit_conversion import unit_conversion
import os, sys, json, logging

class ElkContext(object):
    """context for elk parser"""

    def startedParsing(self, path, parser):
      """called when parsing starts"""
      self.parser = parser
      self.secMethodIndex = None
      self.secSystemIndex = None
      self.enTot = []
      self.atom_pos = []
      self.atom_labels = []
      self.spinTreat = None

    def onOpen_section_system(self, backend, gIndex, section):
     self.secSystemIndex = gIndex

    def onOpen_section_method(self, backend, gIndex, section):
      self.secMethodIndex = gIndex

    def onClose_x_elk_section_lattice_vectors(self, backend, gIndex, section):
      latticeX = section["x_elk_geometry_lattice_vector_x"]
      latticeY = section["x_elk_geometry_lattice_vector_y"]
      latticeZ = section["x_elk_geometry_lattice_vector_z"]
      cell = [[latticeX[0],latticeY[0],latticeZ[0]],
              [latticeX[1],latticeY[1],latticeZ[1]],
              [latticeX[2],latticeY[2],latticeZ[2]]]
      backend.addValue("simulation_cell", cell)

    def onClose_x_elk_section_reciprocal_lattice_vectors(self, backend, gIndex, section):
      recLatticeX = section["x_elk_geometry_reciprocal_lattice_vector_x"]
      recLatticeY = section["x_elk_geometry_reciprocal_lattice_vector_y"]
      recLatticeZ = section["x_elk_geometry_reciprocal_lattice_vector_z"]
      recCell = [[recLatticeX[0],recLatticeY[0],recLatticeZ[0]],
              [recLatticeX[1],recLatticeY[1],recLatticeZ[1]],
              [recLatticeX[2],recLatticeY[2],recLatticeZ[2]]]
      backend.addValue("x_elk_simulation_reciprocal_cell", recCell)

    def onClose_x_elk_section_xc(self, backend, gIndex, section):
      xcNr = section["x_elk_xc_functional"][0]
      xc_internal_map = {
        2: ['LDA_C_PZ', 'LDA_X_PZ'],
        3: ['LDA_C_PW', 'LDA_X_PZ'],
        4: ['LDA_C_XALPHA'],
        5: ['LDA_C_VBH'],
        20: ['GGA_C_PBE', 'GGA_X_PBE'],
        21: ['GGA_C_PBE', 'GGA_X_PBE_R'],
        22: ['GGA_C_PBE_SOL', 'GGA_X_PBE_SOL'],
        26: ['GGA_C_PBE', 'GGA_X_WC'],
        30: ['GGA_C_AM05', 'GGA_X_AM05']
        }
      for xcName in xc_internal_map[xcNr]:
        gi = backend.openSection("section_XC_functionals")
        backend.addValue("XC_functional_name", xcName)
        backend.closeSection("section_XC_functionals", gi)

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
      backend.addValue('single_configuration_to_calculation_method_ref', self.secMethodIndex)
      backend.addValue('single_configuration_calculation_to_system_ref', self.secSystemIndex)
      dirPath = os.path.dirname(self.parser.fIn.name)
      dosFile = os.path.join(dirPath, "TDOS.OUT")
      eigvalFile = os.path.join(dirPath, "EIGVAL.OUT")
      if os.path.exists(dosFile):
        dosGIndex=backend.openSection("section_dos")
        with open(dosFile) as f:
            dosE=[]
            dosV=[]
            fromH = unit_conversion.convert_unit_function("hartree", "J")
            while True:
                line = f.readline()
                if not line: break
                nrs = list(map(float,line.split()))
                if len(nrs) == 2:
                    dosV.append(nrs[1])
                    dosE.append(fromH(nrs[0]))
                elif len(nrs) != 0:
                    raise Exception("Found more than two values in dos file %s" % dosFile)
            backend.addArrayValues("dos_values", np.asarray(dosV))
            backend.addArrayValues("dos_energies", np.asarray(dosE))
        backend.closeSection("section_dos", dosGIndex)
      if os.path.exists(eigvalFile):
        eigvalGIndex = backend.openSection("section_eigenvalues")
        with open(eigvalFile) as g:
            eigvalKpoint=[]
            eigvalVal=[]
            eigvalOcc=[]
            eigvalValSpin = [[],[]]
            eigvalOccSpin = [[],[]]
            fromH = unit_conversion.convert_unit_function("hartree", "J")
            while 1:
              s = g.readline()
              if not s: break
              s = s.strip()
              if len(s) < 20:
                if "nstsv" in s.split():
                  nstsv = int(s.split()[0])
                  nstsv2=int(nstsv/2)
                elif "nkpt" in s.split():
                  nkpt = int(s.split()[0])
                continue
              elif len(s) > 50:
                eigvalVal.append([])
                eigvalOcc.append([])
                eigvalKpoint.append(list(map(float, s.split()[1:4])))
              else:
                try: int(s[0])
                except ValueError:
                  continue
                else:
                  n, e, occ = s.split()
                  eigvalVal[-1].append(fromH(float(e)))
                  eigvalOcc[-1].append(float(occ))
            if not self.spinTreat:
              backend.addArrayValues("eigenvalues_values", np.asarray([eigvalVal]))
              backend.addArrayValues("eigenvalues_occupation", np.asarray([eigvalOcc]))
            else:
              for i in range(0,nkpt):
                eigvalValSpin[0].append(eigvalVal[i][0:nstsv2])
                eigvalOccSpin[0].append(eigvalOcc[i][0:nstsv2])
                eigvalValSpin[1].append(eigvalVal[i][nstsv2:nstsv])
                eigvalOccSpin[1].append(eigvalOcc[i][nstsv2:nstsv])
              backend.addArrayValues("eigenvalues_values", np.asarray(eigvalValSpin))
              backend.addArrayValues("eigenvalues_occupation", np.asarray(eigvalOccSpin))
            backend.addArrayValues("eigenvalues_kpoints", np.asarray(eigvalKpoint))
            backend.closeSection("section_eigenvalues",eigvalGIndex)
#            backend.addArrayValues("eigenvalues_kpoints", np.asarray(eigvalKpoint))
#            backend.addArrayValues("eigenvalues_values", np.asarray([eigvalVal]))
#            backend.addArrayValues("eigenvalues_occupation", np.asarray([eigvalOcc]))
#            backend.closeSection("section_eigenvalues",eigvalGIndex)
      backend.addValue("energy_total", self.enTot[-1])

    def onClose_x_elk_section_spin(self, backend, gIndex, section):
      spin = section["x_elk_spin_treatment"][0]
      spin = spin.strip()
      if spin == "spin-polarised":
        self.spinTreat = True
      else:
        self.spinTreat = False

    def onClose_section_system(self, backend, gIndex, section):
      backend.addArrayValues('configuration_periodic_dimensions', np.asarray([True, True, True]))
      self.secSystemDescriptionIndex = gIndex
      self.secSystemDescriptionIndex = gIndex

      if self.atom_pos:
         backend.addArrayValues('atom_positions', np.asarray(self.atom_pos))
      self.atom_pos = []
      if self.atom_labels is not None:
         backend.addArrayValues('atom_labels', np.asarray(self.atom_labels))
      self.atom_labels = []
    def onClose_x_elk_section_atoms_group(self, backend, gIndex, section):
      pos = [section['x_elk_geometry_atom_positions_' + i] for i in ['x', 'y', 'z']]
      pl = [len(comp) for comp in pos]
      natom = pl[0]
      if pl[1] != natom or pl[2] != natom:
        raise Exception("invalid number of atoms in various components %s" % pl)
      for i in range(natom):
        self.atom_pos.append([pos[0][i], pos[1][i], pos[2][i]])
      self.atom_labels = self.atom_labels + (section['x_elk_geometry_atom_labels'] * natom)

    def onClose_section_scf_iteration(self, backend, gIndex, section):
      Etot = section["energy_total_scf_iteration"]
      self.enTot.append(Etot[0])

# description of the input
mainFileDescription = \
    SM(name = "root matcher",
       startReStr = "",
       weak = True,
       subMatchers = [
         SM(name = "header",
         startReStr = r"\s*\|\s*Elk version\s*(?P<program_version>[-a-zA-Z0-9\.]+)\s*started\s*",
         fixedStartValues={'program_name': 'elk', 'program_basis_set_type': '(L)APW+lo' },
            sections = ["section_run", "section_method"],
         subMatchers = [
           SM(name = 'input',
              startReStr = r"\|\sGround-state\s*[-a-zA-Z\s]+\s*\|\s",
              endReStr = r"\|\sDensity and potential initialised from atomic data\s",
              sections = ['section_system'],
              subMatchers = [
                SM(startReStr = r"\s*Lattice vectors :",
                sections = ["x_elk_section_lattice_vectors"],
                subMatchers = [

    SM(startReStr = r"\s*(?P<x_elk_geometry_lattice_vector_x__bohr>[-+0-9.]+)\s+(?P<x_elk_geometry_lattice_vector_y__bohr>[-+0-9.]+)\s+(?P<x_elk_geometry_lattice_vector_z__bohr>[-+0-9.]+)", repeats = True)
                ]),
                SM(startReStr = r"Reciprocal lattice vectors :",
                sections = ["x_elk_section_reciprocal_lattice_vectors"],
                subMatchers = [

    SM(startReStr = r"\s*(?P<x_elk_geometry_reciprocal_lattice_vector_x__bohr_1>[-+0-9.]+)\s+(?P<x_elk_geometry_reciprocal_lattice_vector_y__bohr_1>[-+0-9.]+)\s+(?P<x_elk_geometry_reciprocal_lattice_vector_z__bohr_1>[-+0-9.]+)", repeats = True)
                ]),
    SM(r"\s*Unit cell volume\s*:\s*(?P<x_elk_unit_cell_volume__bohr3>[-0-9.]+)"),
    SM(r"\s*Brillouin zone volume\s*:\s*(?P<x_elk_brillouin_zone_volume__bohr_3>[-0-9.]+)"),
    SM(r"\s*Species\s*:\s*[0-9]\s*\((?P<x_elk_geometry_atom_labels>[-a-zA-Z0-9]+)\)", repeats = True,
      sections = ["x_elk_section_atoms_group"],
       subMatchers = [
    SM(r"\s*muffin-tin radius\s*:\s*(?P<x_elk_muffin_tin_radius__bohr>[-0-9.]+)", repeats = True),
    SM(r"\s*number of radial points in muffin-tin\s*:\s*(?P<x_elk_muffin_tin_points>[-0-9.]+)", repeats = True),
    SM(startReStr = r"\s*atomic positions\s*\(lattice\)\, magnetic fields \(Cartesian\)\s*:\s*",
      subMatchers = [
        SM(r"\s*(?P<x_elk_geometry_atom_number>[+0-9]+)\s*:\s*(?P<x_elk_geometry_atom_positions_x__bohr>[-+0-9.]+)\s*(?P<x_elk_geometry_atom_positions_y__bohr>[-+0-9.]+)\s*(?P<x_elk_geometry_atom_positions_z__bohr>[-+0-9.]+)", repeats = True)
      ])
    ]),
    SM(startReStr = r"\s*Spin treatment\s*:\s*",
      sections = ["x_elk_section_spin"],
        subMatchers = [
        SM(r"\s*(?P<x_elk_spin_treatment>[-a-zA-Z\s*]+)")
     ]),
    SM(r"\s*k-point grid\s*:\s*(?P<x_elk_number_kpoint_x>[-0-9.]+)\s+(?P<x_elk_number_kpoint_y>[-0-9.]+)\s+(?P<x_elk_number_kpoint_z>[-0-9.]+)"),
    SM(r"\s*k-point offset\s*:\s*(?P<x_elk_kpoint_offset_x>[-0-9.]+)\s+(?P<x_elk_kpoint_offset_y>[-0-9.]+)\s+(?P<x_elk_kpoint_offset_z>[-0-9.]+)"),
    SM(r"\s*Total number of k-points\s*:\s*(?P<x_elk_number_kpoints>[-0-9.]+)"),
    SM(r"\s*Muffin-tin radius times maximum \|G\+k\|\s*:\s*(?P<x_elk_rgkmax__bohr>[-0-9.]+)"),
    SM(r"\s*Maximum \|G\+k\| for APW functions\s*:\s*(?P<x_elk_gkmax__bohr_1>[-0-9.]+)"),
    SM(r"\s*Maximum \|G\| for potential and density\s*:\s*(?P<x_elk_gmaxvr__bohr_1>[-0-9.]+)"),
    SM(r"\s*G-vector grid sizes\s*:\s*(?P<x_elk_gvector_size_x>[-0-9.]+)\s+(?P<x_elk_gvector_size_y>[-0-9.]+)\s+(?P<x_elk_gvector_size_z>[-0-9.]+)"),
    SM(r"\s*Number of G-vectors\s*:\s*(?P<x_elk_gvector_total>[-0-9.]+)"),
    SM(startReStr = r"\s*Maximum angular momentum used for\s*",
        subMatchers = [
          SM(r"\s*APW functions\s*:\s*(?P<x_elk_lmaxapw>[-0-9.]+)")
        ]),
    SM(r"\s*Total nuclear charge\s*:\s*(?P<x_elk_nuclear_charge>[-0-9.]+)"),
    SM(r"\s*Total core charge\s*:\s*(?P<x_elk_core_charge>[-0-9.]+)"),
    SM(r"\s*Total valence charge\s*:\s*(?P<x_elk_valence_charge>[-0-9.]+)"),
    SM(r"\s*Total electronic charge\s*:\s*(?P<x_elk_electronic_charge>[-0-9.]+)"),
    SM(r"\s*Effective Wigner radius, r_s\s*:\s*(?P<x_elk_wigner_radius>[-0-9.]+)"),
    SM(r"\s*Number of empty states\s*:\s*(?P<x_elk_empty_states>[-0-9.]+)"),
    SM(r"\s*Total number of valence states\s*:\s*(?P<x_elk_valence_states>[-0-9.]+)"),
    SM(r"\s*Total number of core states\s*:\s*(?P<x_elk_core_states>[-0-9.]+)"),
    SM(r"\s*Total number of local-orbitals\s*:\s*(?P<x_elk_lo>[-0-9.]+)"),
    SM(r"\s*Smearing width\s*:\s*(?P<x_elk_smearing_width__hartree>[-0-9.]+)"),
    SM(startReStr = r"\s*Exchange-correlation functional\s*:\s*(?P<x_elk_xc_functional>[-0-9.]+)",
       sections = ['x_elk_section_xc'])
           ]),
            SM(name = "single configuration iteration",
              startReStr = r"\|\s*Self-consistent loop started\s*\|",
              sections = ["section_single_configuration_calculation"],
              repeats = True,
              subMatchers = [
                SM(name = "scfi totE",
                 startReStr =r"\|\s*[-a-zA-Z]+ number\s*:",
                  sections = ["section_scf_iteration"],
                  repeats = True,
                  subMatchers = [
                   SM(r"\s*Fermi\s*:\s*(?P<x_elk_fermi_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*sum of eigenvalues\s*:\s*(?P<energy_sum_eigenvalues_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*electron kinetic\s*:\s*(?P<electronic_kinetic_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*core electron kinetic\s*:\s*(?P<x_elk_core_electron_kinetic_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Coulomb\s*:\s*(?P<x_elk_coulomb_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Coulomb potential\s*:\s*(?P<x_elk_coulomb_potential_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*nuclear-nuclear\s*:\s*(?P<x_elk_nuclear_nuclear_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*electron-nuclear\s*:\s*(?P<x_elk_electron_nuclear_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Hartree\s*:\s*(?P<x_elk_hartree_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Madelung\s*:\s*(?P<x_elk_madelung_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*xc potential\s*:\s*(?P<energy_XC_potential_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*exchange\s*:\s*(?P<x_elk_exchange_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*correlation\s*:\s*(?P<x_elk_correlation_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*electron entropic\s*:\s*(?P<x_elk_electron_entropic_energy_scf_iteration__hartree>[-0-9.]+([E]?[-]?[0-9]+))"),
                   SM(r"\s*total energy\s*:\s*(?P<energy_total_scf_iteration__hartree>[-0-9.]+([E]?[-]?[0-9]+))"),
                   SM(r"\s*Density of states at Fermi energy\s*:\s*(?P<x_elk_dos_fermi_scf_iteration__hartree_1>[-0-9.]+)"),
                   SM(r"\s*Estimated indirect band gap\s*:\s*(?P<x_elk_indirect_gap_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Estimated direct band gap\s*:\s*(?P<x_elk_direct_gap_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*core\s*:\s*(?P<x_elk_core_charge_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*valence\s*:\s*(?P<x_elk_valence_charge_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*interstitial\s*:\s*(?P<x_elk_interstitial_charge_scf_iteration>[-0-9.]+)"),
                  ]) #,
               ]
            )
          ])
    ])

parserInfo = {
  "name": "Elk"
}

# metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/elk.nomadmetainfo.json"))
# metaInfoEnv, warnings = loadJsonFile(filePath = metaInfoPath, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, uri = None)

cachingLevelForMetaName = {
                            "x_elk_geometry_lattice_vector_x":CachingLevel.Cache,
                            "x_elk_geometry_lattice_vector_y":CachingLevel.Cache,
                            "x_elk_geometry_lattice_vector_z":CachingLevel.Cache,
                            "x_elk_section_lattice_vectors": CachingLevel.Ignore,
                            "x_elk_geometry_reciprocal_lattice_vector_x":CachingLevel.Cache,
                            "x_elk_geometry_reciprocal_lattice_vector_y":CachingLevel.Cache,
                            "x_elk_geometry_reciprocal_lattice_vector_z":CachingLevel.Cache,
                            "x_elk_section_reciprocal_lattice_vectors": CachingLevel.Ignore
                          }


class ElkParser():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
       from unittest.mock import patch
       logging.info('elk parser started')
       logging.getLogger('nomadcore').setLevel(logging.WARNING)
       backend = self.backend_factory("elk.nomadmetainfo.json")
       with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
           mainFunction(
               mainFileDescription=mainFileDescription,
               metaInfoEnv=None,
               parserInfo = parserInfo,
               cachingLevelForMetaName = cachingLevelForMetaName,
               superContext=ElkContext(),
               superBackend=backend,
               defaultSectionCachingLevel = True)
       return backend



import numpy as np
from bitarray import bitarray

def match_fingerprints(fingerprint1, fingerprint2):
    if fingerprint1.grid_id != fingerprint2.grid_id:
        raise AssertionError('Can not calculate similarity of fingerprints that have been calculated with different grids.')
    num_bins = int(fingerprint1.grid_id.split(':')[1])
    fp1 = bitarray()
    fp2 = bitarray()
    fp1.frombytes(bytes.fromhex(fingerprint1.bins))
    fp2.frombytes(bytes.fromhex(fingerprint2.bins))
    start_index = max([fingerprint1.indices[0], fingerprint2.indices[0]])
    stop_index = min([fingerprint1.indices[1], fingerprint2.indices[1]])
    # find offsets
    dsp1 = (start_index - fingerprint1.indices[0]) * num_bins
    dsp2 = (start_index - fingerprint2.indices[0]) * num_bins
    dep1 = (fingerprint1.indices[1] - stop_index) * num_bins
    dep2 = (fingerprint2.indices[1] - stop_index) * num_bins
    fp1 = fp1[dsp1:len(fp1) - 1 - dep1]
    fp2 = fp2[dsp2:len(fp2) - 1 - dep2]
    return fp1, fp2

def tanimoto_similarity(fingerprint1, fingerprint2):
    fp1, fp2 = match_fingerprints(fingerprint1, fingerprint2)
    a = fp1.count()
    b = fp2.count()
    c = (fp1 & fp2).count()
    try:
        tc = c / float(a + b - c)
    except ZeroDivisionError:
        tc = 0
    return tc



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os.path

from nomad import config, datamodel

from .parser import MissingParser, BrokenParser, Parser, ArchiveParser
from .legacy import LegacyParser, VaspOutcarParser
from .artificial import EmptyParser, GenerateRandomParser, TemplateParser, ChaosParser

from eelsdbconverter import EELSApiJsonConverter
from mpesparser import MPESParser
from aptfimparser import APTFIMParser
from vaspparser import VASPParser
from phonopyparser import PhonopyParser
from elasticparser import ElasticParser
from lammpsparser import LammpsParser

try:
    # these packages are not available without parsing extra, which is ok, if the
    # parsers are only initialized to load their metainfo definitions
    import magic
    import gzip
    import bz2
    import lzma

    _compressions = {
        b'\x1f\x8b\x08': ('gz', gzip.open),
        b'\x42\x5a\x68': ('bz2', bz2.open),
        b'\xfd\x37\x7a': ('xz', lzma.open)
    }

    encoding_magic = magic.Magic(mime_encoding=True)

except ImportError:
    pass


def match_parser(mainfile_path: str, strict=True) -> Parser:
    '''
    Performs parser matching. This means it take the given mainfile and potentially
    opens it with the given callback and tries to identify a parser that can parse
    the file.

    This is determined by filename (e.g. *.out), mime type (e.g. text/*, application/xml),
    and beginning file contents.

    Arguments:
        mainfile_path: Path to the mainfile
        strict: Only match strict parsers, e.g. no artificial parsers for missing or empty entries.

    Returns: The parser, or None if no parser could be matched.
    '''
    mainfile = os.path.basename(mainfile_path)
    if mainfile.startswith('.') or mainfile.startswith('~'):
        return None

    with open(mainfile_path, 'rb') as f:
        compression, open_compressed = _compressions.get(f.read(3), (None, open))

    with open_compressed(mainfile_path, 'rb') as cf:  # type: ignore
        buffer = cf.read(config.parser_matching_size)

    mime_type = magic.from_buffer(buffer, mime=True)

    decoded_buffer = None
    encoding = None
    try:  # Try to open the file as a string for regex matching.
        decoded_buffer = buffer.decode('utf-8')
    except UnicodeDecodeError:
        # This file is either binary or has wrong encoding
        encoding = encoding_magic.from_buffer(buffer)

        if config.services.force_raw_file_decoding:
            encoding = 'iso-8859-1'

        if encoding in ['iso-8859-1']:
            try:
                decoded_buffer = buffer.decode(encoding)
            except Exception:
                pass

    for parser in parsers:
        if strict and isinstance(parser, (MissingParser, EmptyParser)):
            continue

        if parser.is_mainfile(mainfile_path, mime_type, buffer, decoded_buffer, compression):
            # potentially convert the file
            if encoding in ['iso-8859-1']:
                try:
                    with open(mainfile_path, 'rb') as binary_file:
                        content = binary_file.read().decode(encoding)
                except Exception:
                    pass
                else:
                    with open(mainfile_path, 'wt') as text_file:
                        text_file.write(content)

            # TODO: deal with multiple possible parser specs
            return parser

    return None


parsers = [
    GenerateRandomParser(),
    TemplateParser(),
    ChaosParser(),
    PhonopyParser(),
    VASPParser(),
    VaspOutcarParser(
        name='parsers/vasp-outcar', code_name='VASP', code_homepage='https://www.vasp.at/',
        parser_class_name='vaspparser.VaspOutcarParser',
        mainfile_name_re=r'(.*/)?OUTCAR(\.[^\.]*)?',
        mainfile_contents_re=(r'^\svasp\.')
    ),
    LegacyParser(
        name='parsers/exciting', code_name='exciting', code_homepage='http://exciting-code.org/',
        parser_class_name='excitingparser.ExcitingParser',
        mainfile_name_re=r'^.*.OUT(\.[^/]*)?$',
        mainfile_contents_re=(r'EXCITING.*started')
    ),
    LegacyParser(
        name='parsers/fhi-aims', code_name='FHI-aims', code_homepage='https://aimsclub.fhi-berlin.mpg.de/',
        parser_class_name='fhiaimsparser.FHIaimsParser',
        mainfile_contents_re=(
            r'^(.*\n)*'
            r'?\s*Invoking FHI-aims \.\.\.'
            # r'?\s*Version'
        )
    ),
    LegacyParser(
        name='parsers/cp2k', code_name='CP2K', code_homepage='https://www.cp2k.org/',
        parser_class_name='cp2kparser.CP2KParser',
        mainfile_contents_re=(
            r'\*\*\*\* \*\*\*\* \*\*\*\*\*\*  \*\*  PROGRAM STARTED AT\s.*\n'
            r' \*\*\*\*\* \*\* \*\*\*  \*\*\* \*\*   PROGRAM STARTED ON\s*.*\n'
            r' \*\*    \*\*\*\*   \*\*\*\*\*\*    PROGRAM STARTED BY .*\n'
            r' \*\*\*\*\* \*\*    \*\* \*\* \*\*   PROGRAM PROCESS ID .*\n'
            r'  \*\*\*\* \*\*  \*\*\*\*\*\*\*  \*\*  PROGRAM STARTED IN .*\n'
        )
    ),
    LegacyParser(
        name='parsers/crystal', code_name='Crystal', code_homepage='https://www.crystal.unito.it/',
        parser_class_name='crystalparser.CrystalParser',
        mainfile_contents_re=(
            r'(CRYSTAL\s*\n\d+ \d+ \d+)|(CRYSTAL will run on \d+ processors)|'
            r'(\s*\*\s*CRYSTAL[\d]+\s*\*\s*\*\s*(public|Release) \: [\d\.]+.*\*)|'
            r'(Executable:\s*[/_\-a-zA-Z0-9]*MPPcrystal)'
        )
    ),
    # The main contents regex of CPMD was causing a catostrophic backtracking issue
    # when searching through the first 500 bytes of main files. We decided
    # to use only a portion of the regex to avoid that issue.
    LegacyParser(
        name='parsers/cpmd', code_name='CPMD', code_homepage='https://www.lcrc.anl.gov/for-users/software/available-software/cpmd/',
        parser_class_name='cpmdparser.CPMDParser',
        mainfile_contents_re=(
            # r'\s+\*\*\*\*\*\*  \*\*\*\*\*\*    \*\*\*\*  \*\*\*\*  \*\*\*\*\*\*\s*'
            # r'\s+\*\*\*\*\*\*\*  \*\*\*\*\*\*\*   \*\*\*\*\*\*\*\*\*\*  \*\*\*\*\*\*\*\s+'
            r'\*\*\*       \*\*   \*\*\*  \*\* \*\*\*\* \*\*  \*\*   \*\*\*'
            # r'\s+\*\*        \*\*   \*\*\*  \*\*  \*\*  \*\*  \*\*    \*\*\s+'
            # r'\s+\*\*        \*\*\*\*\*\*\*   \*\*      \*\*  \*\*    \*\*\s+'
            # r'\s+\*\*\*       \*\*\*\*\*\*    \*\*      \*\*  \*\*   \*\*\*\s+'
            # r'\s+\*\*\*\*\*\*\*  \*\*        \*\*      \*\*  \*\*\*\*\*\*\*\s+'
            # r'\s+\*\*\*\*\*\*  \*\*        \*\*      \*\*  \*\*\*\*\*\*\s+'
        )
    ),
    LegacyParser(
        name='parsers/nwchem', code_name='NWChem', code_homepage='http://www.nwchem-sw.org/',
        parser_class_name='nwchemparser.NWChemParser',
        mainfile_contents_re=(
            r'Northwest Computational Chemistry Package \(NWChem\) (\d+\.)+\d+'
        )
    ),
    LegacyParser(
        name='parsers/bigdft', code_name='BigDFT', code_homepage='http://bigdft.org/',
        parser_class_name='bigdftparser.BigDFTParser',
        mainfile_contents_re=(
            # r'__________________________________ A fast and precise DFT wavelet code\s*'
            # r'\|     \|     \|     \|     \|     \|\s*'
            # r'\|     \|     \|     \|     \|     \|      BBBB         i       gggggg\s*'
            # r'\|_____\|_____\|_____\|_____\|_____\|     B    B               g\s*'
            # r'\|     \|  :  \|  :  \|     \|     \|    B     B        i     g\s*'
            # r'\|     \|-0\+--\|-0\+--\|     \|     \|    B    B         i     g        g\s*'
            r'\|_____\|__:__\|__:__\|_____\|_____\|___ BBBBB          i     g         g\s*'
            # r'\|  :  \|     \|     \|  :  \|     \|    B    B         i     g         g\s*'
            # r'\|--\+0-\|     \|     \|-0\+--\|     \|    B     B     iiii     g         g\s*'
            # r'\|__:__\|_____\|_____\|__:__\|_____\|    B     B        i      g        g\s*'
            # r'\|     \|  :  \|  :  \|     \|     \|    B BBBB        i        g      g\s*'
            # r'\|     \|-0\+--\|-0\+--\|     \|     \|    B        iiiii          gggggg\s*'
            # r'\|_____\|__:__\|__:__\|_____\|_____\|__BBBBB\s*'
            # r'\|     \|     \|     \|  :  \|     \|                           TTTTTTTTT\s*'
            # r'\|     \|     \|     \|--\+0-\|     \|  DDDDDD          FFFFF        T\s*'
            # r'\|_____\|_____\|_____\|__:__\|_____\| D      D        F        TTTT T\s*'
            # r'\|     \|     \|     \|  :  \|     \|D        D      F        T     T\s*'
            # r'\|     \|     \|     \|--\+0-\|     \|D         D     FFFF     T     T\s*'
            # r'\|_____\|_____\|_____\|__:__\|_____\|D___      D     F         T    T\s*'
            # r'\|     \|     \|  :  \|     \|     \|D         D     F          TTTTT\s*'
            # r'\|     \|     \|--\+0-\|     \|     \| D        D     F         T    T\s*'
            # r'\|_____\|_____\|__:__\|_____\|_____\|          D     F        T     T\s*'
            # r'\|     \|     \|     \|     \|     \|         D               T    T\s*'
            # r'\|     \|     \|     \|     \|     \|   DDDDDD       F         TTTT\s*'
            # r'\|_____\|_____\|_____\|_____\|_____\|______                    www\.bigdft\.org'
        )
    ),
    LegacyParser(
        name='parsers/wien2k', code_name='WIEN2k', code_homepage='http://www.wien2k.at/',
        parser_class_name='wien2kparser.Wien2kParser',
        mainfile_contents_re=r'\s*---------\s*:ITE[0-9]+:\s*[0-9]+\.\s*ITERATION\s*---------'
    ),
    LegacyParser(
        name='parsers/band', code_name='BAND', code_homepage='https://www.scm.com/product/band_periodicdft/',
        parser_class_name='bandparser.BANDParser',
        mainfile_contents_re=r' +\* +Amsterdam Density Functional +\(ADF\)'),
    LegacyParser(
        name='parsers/gaussian', code_name='Gaussian', code_homepage='http://gaussian.com/',
        parser_class_name='gaussianparser.GaussianParser',
        mainfile_mime_re=r'.*',
        mainfile_contents_re=(
            r'\s*Cite this work as:'
            r'\s*Gaussian [0-9]+, Revision [A-Za-z0-9\.]*,')
    ),
    LegacyParser(
        name='parsers/quantumespresso', code_name='Quantum Espresso', code_homepage='https://www.quantum-espresso.org/',
        parser_class_name='quantumespressoparser.QuantumEspressoParserPWSCF',
        mainfile_contents_re=(
            r'(Program PWSCF.*starts)|'
            r'(Current dimensions of program PWSCF are)')
        #    r'^(.*\n)*'
        #    r'\s*Program (\S+)\s+v\.(\S+)(?:\s+\(svn\s+rev\.\s+'
        #    r'(\d+)\s*\))?\s+starts[^\n]+'
        #    r'(?:\s*\n?)*This program is part of the open-source Quantum')
    ),
    LegacyParser(
        name='parsers/abinit', code_name='ABINIT', code_homepage='https://www.abinit.org/',
        parser_class_name='abinitparser.AbinitParser',
        mainfile_contents_re=(r'^\n*\.Version\s*[0-9.]*\s*of ABINIT\s*')
    ),
    LegacyParser(
        name='parsers/orca', code_name='ORCA', code_homepage='https://orcaforum.kofo.mpg.de/',
        parser_class_name='orcaparser.OrcaParser',
        mainfile_contents_re=(
            r'\s+\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
            r'\s+\* O   R   C   A \*\s*'
            r'\s+\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
            r'\s*'
            r'\s*--- An Ab Initio, DFT and Semiempirical electronic structure package ---\s*')
    ),
    LegacyParser(
        name='parsers/castep', code_name='CASTEP', code_homepage='http://www.castep.org/',
        parser_class_name='castepparser.CastepParser',
        mainfile_contents_re=(r'\s\|\s*CCC\s*AA\s*SSS\s*TTTTT\s*EEEEE\s*PPPP\s*\|\s*')
    ),
    LegacyParser(
        name='parsers/dl-poly', code_name='DL_POLY', code_homepage='https://www.scd.stfc.ac.uk/Pages/DL_POLY.aspx',
        parser_class_name='dlpolyparser.DlPolyParserWrapper',
        mainfile_contents_re=(r'\*\* DL_POLY \*\*')
    ),
    LegacyParser(
        name='parsers/lib-atoms', code_name='libAtoms', code_homepage='https://libatoms.github.io/',
        parser_class_name='libatomsparser.LibAtomsParserWrapper',
        mainfile_contents_re=(r'\s*<GAP_params\s')
    ),
    LegacyParser(
        name='parsers/octopus', code_name='Octopus', code_homepage='https://octopus-code.org/',
        parser_class_name='octopusparser.OctopusParserWrapper',
        mainfile_contents_re=(r'\|0\) ~ \(0\) \|')
        # We decided to use the octopus eyes instead of
        # r'\*{32} Grid \*{32}Simulation Box:' since it was so far down in the file.
    ),
    # match gpaw2 first, other .gpw files are then considered to be "gpaw1"
    LegacyParser(
        name='parsers/gpaw2', code_name='GPAW', code_homepage='https://wiki.fysik.dtu.dk/gpaw/',
        parser_class_name='gpawparser.GPAWParser2Wrapper',
        mainfile_binary_header=b'GPAW',
        mainfile_name_re=(r'^.*\.(gpw2|gpw)$'),
        mainfile_mime_re=r'application/(x-tar|octet-stream)'
    ),
    LegacyParser(
        name='parsers/gpaw', code_name='GPAW', code_homepage='https://wiki.fysik.dtu.dk/gpaw/',
        parser_class_name='gpawparser.GPAWParserWrapper',
        mainfile_name_re=(r'^.*\.gpw$'),
        mainfile_mime_re=r'application/(x-tar|octet-stream)'
    ),
    LegacyParser(
        name='parsers/atk', code_name='ATK', code_homepage='https://www.synopsys.com/silicon/quantumatk.html',
        parser_class_name='atkparser.ATKParserWrapper',
        # mainfile_contents_re=r'',  # We can't read .gpw as txt - of UlmGPAW|AFFormatGPAW'
        mainfile_name_re=r'^.*\.nc',
        # The previously used mime type r'application/x-netcdf' wasn't found by magic library.
        mainfile_mime_re=r'application/octet-stream'
    ),
    LegacyParser(
        name='parsers/gulp', code_name='gulp', code_homepage='http://gulp.curtin.edu.au/gulp/',
        parser_class_name='gulpparser.GULPParser',
        mainfile_contents_re=(
            r'\s*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*'
            r'\*\*\*\*\*\*\*\*\*\*\*\*\*\s*'
            r'\s*\*\s*GENERAL UTILITY LATTICE PROGRAM\s*\*\s*')
    ),
    LegacyParser(
        name='parsers/siesta', code_name='Siesta', code_homepage='https://departments.icmab.es/leem/siesta/',
        parser_class_name='siestaparser.SiestaParser',
        mainfile_contents_re=(
            r'(Siesta Version: siesta-|SIESTA [0-9]\.[0-9]\.[0-9])|'
            r'(\*\s*WELCOME TO SIESTA\s*\*)')
    ),
    LegacyParser(
        name='parsers/elk', code_name='elk', code_homepage='http://elk.sourceforge.net/',
        parser_class_name='elkparser.ElkParser',
        mainfile_contents_re=r'\| Elk version [0-9.a-zA-Z]+ started \|'
    ),
    ElasticParser(),
    LegacyParser(
        name='parsers/gamess', code_name='GAMESS', code_homepage='https://www.msg.chem.iastate.edu/gamess/versions.html',
        parser_class_name='gamessparser.GamessParser',
        mainfile_contents_re=(
            r'\s*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
            r'\s*\*\s*GAMESS VERSION =\s*(.*)\*\s*'
            r'\s*\*\s*FROM IOWA STATE UNIVERSITY\s*\*\s*')
    ),
    LegacyParser(
        name='parsers/turbomole', code_name='turbomole', code_homepage='https://www.turbomole.org/',
        parser_class_name='turbomoleparser.TurbomoleParser',
        mainfile_contents_re=(
            r'Copyright \(C\) [0-9]+ TURBOMOLE GmbH, Karlsruhe')
    ),
    MPESParser(),
    APTFIMParser(),
    EELSApiJsonConverter(),
    LegacyParser(
        name='parsers/qbox', code_name='qbox', code_homepage='http://qboxcode.org/', domain='dft',
        parser_class_name='qboxparser.QboxParser',
        mainfile_mime_re=r'(application/xml)|(text/.*)',
        mainfile_contents_re=(r'http://qboxcode.org')
    ),
    LegacyParser(
        name='parsers/dmol', code_name='DMol3', code_homepage='http://dmol3.web.psi.ch/dmol3.html', domain='dft',
        parser_class_name='dmol3parser.Dmol3Parser',
        mainfile_name_re=r'.*\.outmol',
        mainfile_contents_re=r'Materials Studio DMol\^3'
    ),
    LegacyParser(
        name='parsers/fleur', code_name='fleur', code_homepage='https://www.flapw.de/', domain='dft',
        parser_class_name='fleurparser.FleurParser',
        mainfile_contents_re=r'This output is generated by fleur.'
    ),
    LegacyParser(
        name='parsers/molcas', code_name='MOLCAS', code_homepage='http://www.molcas.org/', domain='dft',
        parser_class_name='molcasparser.MolcasParser',
        mainfile_contents_re=r'M O L C A S'
    ),
    LegacyParser(
        name='parsers/onetep', code_name='ONETEP', code_homepage='https://www.onetep.org/', domain='dft',
        parser_class_name='onetepparser.OnetepParser',
        mainfile_contents_re=r'####### #     # ####### ####### ####### ######'
    ),
    LegacyParser(
        name='parsers/openkim', code_name='OpenKIM', domain='dft',
        parser_class_name='openkimparser.OpenKIMParser',
        mainfile_contents_re=r'OPENKIM'
    ),
    LegacyParser(
        name='parsers/tinker', code_name='TINKER', domain='dft',
        parser_class_name='tinkerparser.TinkerParser',
        mainfile_contents_re=r'TINKER  ---  Software Tools for Molecular Design'
    ),
    LammpsParser(),
    LegacyParser(
        name='parsers/amber', code_name='Amber', domain='dft',
        parser_class_name='amberparser.AMBERParser',
        mainfile_contents_re=r'\s*Amber\s[0-9]+\s[A-Z]+\s*[0-9]+'
    ),
    LegacyParser(
        name='parsers/gromacs', code_name='Gromacs', domain='dft',
        parser_class_name='gromacsparser.GROMACSParser',
        mainfile_contents_re=r'GROMACS - gmx mdrun'
    ),
    LegacyParser(
        name='parsers/gromos', code_name='Gromos', domain='dft',
        parser_class_name='gromosparser.GromosParser',
        mainfile_contents_re=r'Bugreports to http://www.gromos.net'
    ),
    LegacyParser(
        name='parsers/namd', code_name='Namd', domain='dft',
        parser_class_name='namdparser.NamdParser',
        mainfile_contents_re=r'\s*Info:\s*NAMD\s*[0-9.]+\s*for\s*',
        mainfile_mime_re=r'text/.*',
    ),
    LegacyParser(
        name='parsers/charmm', code_name='Charmm', domain='dft',
        parser_class_name='charmmparser.CharmmParser',
        mainfile_contents_re=r'\s*Chemistry\s*at\s*HARvard\s*Macromolecular\s*Mechanics\s*',
        mainfile_mime_re=r'text/.*',
    ),
    LegacyParser(
        name='parsers/dftbplus', code_name='DFTB+', domain='dft',
        parser_class_name='dftbplusparser.DFTBPlusParser',
        mainfile_contents_re=r'^ Fermi distribution function\s*',
        mainfile_mime_re=r'text/.*',
    ),
    LegacyParser(
        name='parsers/asap', code_name='ASAP', domain='dft',
        parser_class_name='asapparser.AsapParser',
        mainfile_name_re=r'.*.traj$',
        mainfile_mime_re=r'application/octet-stream',
    ),
    LegacyParser(
        name='parsers/fplo', code_name='fplo', domain='dft',
        parser_class_name='fploparser.FploParser',
        mainfile_contents_re=r'\s*\|\s*FULL-POTENTIAL LOCAL-ORBITAL MINIMUM BASIS BANDSTRUCTURE CODE\s*\|\s*',
        mainfile_mime_re=r'text/.*',
    ),
    LegacyParser(
        name='parsers/mopac', code_name='MOPAC', domain='dft',
        parser_class_name='mopacparser.MopacParser',
        mainfile_contents_re=r'\s*\*\*\s*MOPAC\s*([0-9a-zA-Z]*)\s*\*\*\s*',
        mainfile_mime_re=r'text/.*',
    ),
    ArchiveParser()
]

empty_parsers = [
    EmptyParser(
        name='missing/octopus', code_name='Octopus', code_homepage='https://octopus-code.org/',
        domain='dft',
        mainfile_name_re=r'(inp)|(.*/inp)'
    ),
    EmptyParser(
        name='missing/crystal', code_name='Crystal', code_homepage='https://www.crystal.unito.it/index.php',
        domain='dft',
        mainfile_name_re=r'.*\.cryst\.out'
    ),
    EmptyParser(
        name='missing/wien2k', code_name='WIEN2k', code_homepage='http://www.wien2k.at/',
        domain='dft',
        mainfile_name_re=r'.*\.scf'
    ),
    EmptyParser(
        name='missing/fhi-aims', code_name='FHI-aims', code_homepage='https://aimsclub.fhi-berlin.mpg.de/',
        domain='dft',
        mainfile_name_re=r'.*\.fhiaims'
    )
]

if config.use_empty_parsers:
    # There are some entries with PIDs that have mainfiles which do not match what
    # the actual parsers expect. We use the EmptyParser to produce placeholder entries
    # to keep the PIDs. These parsers will not match for new, non migrated data.
    parsers.extend(empty_parsers)

parsers.append(BrokenParser())

''' Instantiation and constructor based config of all parsers. '''

parser_dict = {parser.name: parser for parser in parsers + empty_parsers}  # type: ignore
''' A dict to access parsers by name. Usually 'parsers/<...>', e.g. 'parsers/vasp'. '''

# renamed parsers
parser_dict['parser/broken'] = parser_dict['parsers/broken']
parser_dict['parser/fleur'] = parser_dict['parsers/fleur']
parser_dict['parser/molcas'] = parser_dict['parsers/molcas']
parser_dict['parser/octopus'] = parser_dict['parsers/octopus']
parser_dict['parser/onetep'] = parser_dict['parsers/onetep']

# register code names as possible statistic value to the dft datamodel
code_names = []
for parser in parsers:
    if parser.domain == 'dft' and \
            getattr(parser, 'code_name', None) is not None and \
            getattr(parser, 'code_name') != 'currupted mainfile' and \
            getattr(parser, 'code_name') != 'Template':
        code_names.append(getattr(parser, 'code_name'))
code_names = sorted(set(code_names), key=lambda code_name: code_name.lower())
datamodel.DFTMetadata.code_name.a_search.statistic_values = code_names + [
    config.services.unavailable_value, config.services.not_processed_value]



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from collections import UserDict, OrderedDict
from abc import ABCMeta
import numpy as np

# logging.basicConfig()
logger = logging.getLogger(__name__)


class AtomType(OrderedDict):
    pass


class Atoms:
    def __init__(self):
        self.data = []
        self.atom_id = {}
        self.molecule_id = {}

    def add(self, id, values):
        self.data.append(values)
        self.atom_id[id] = values


    # def __hash__(self):
    #     return hash(self.atom_id)


class AtomStyle(metaclass=ABCMeta):
    pass


class AtomStyleFull(AtomStyle):
    name = 'full'

    # atom-ID molecule-ID atom-type q x y z
    def __init__(self, tom_id, molecule_id, atom_type, q, x, y, z):
        super(AtomStyleFull, self).__init__()

        self.tom_id = tom_id
        self.molecule_id = molecule_id
        self.atom_type = atom_type
        self.q = q
        self.x = x
        self.y = y
        self.z = z


    # @property
    # def x(self):
    #     return self.x


"""TODO
filereader
data numpy? dict? class?

"""

class DataFile(object):
    def __init__(self, filename):
        self.filereader = filename


    @property
    def filereader(self):
        return self._filereader

    @filereader.setter
    def filereader(self, filename):
        self._filereader = filename



if __name__ == '__main__':

    logger.info("Start Datafile module")
    print("hello")

    data = DataFile("data.methene")

    a = Atoms()
    a.add(1,[1,2,3])
    a.add(2,[11,22,33])
    print(a.data)
    print(a.atom_id)

    a.atom_id[1][2]=11111
    print(a.data)
    print(a.atom_id)


    atom_type = AtomType()
    atom_type[2] = {'mass': 1}
    atom_type[1] = {'mass': 11}

    print(atom_type)
























    pass



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import PairStyle


# class None(PairStyle):
#     name = 'none'
#     pass


class Hybrid(PairStyle):
    name = 'hybrid'
    pass


class HybridOverlay(PairStyle):
    name = 'hybrid/overlay'
    pass


class Zero(PairStyle):
    name = 'zero'
    pass


class Adp(PairStyle):
    name = 'adp'
    pass


class Airebo(PairStyle):
    name = 'airebo'
    pass


class AireboMorse(PairStyle):
    name = 'airebo/morse'
    pass


class Beck(PairStyle):
    name = 'beck'
    pass


class Body(PairStyle):
    name = 'body'
    pass


class Bop(PairStyle):
    name = 'bop'
    pass


class Born(PairStyle):
    name = 'born'
    pass


class BornCoulLong(PairStyle):
    name = 'born/coul/long'
    pass


class BornCoulLongCs(PairStyle):
    name = 'born/coul/long/cs'
    pass


class BornCoulMsm(PairStyle):
    name = 'born/coul/msm'
    pass


class BornCoulWolf(PairStyle):
    name = 'born/coul/wolf'
    pass


class Brownian(PairStyle):
    name = 'brownian'
    pass


class BrownianPoly(PairStyle):
    name = 'brownian/poly'
    pass


class Buck(PairStyle):
    name = 'buck'
    pass


class BuckCoulCut(PairStyle):
    name = 'buck/coul/cut'
    pass


class BuckCoulLong(PairStyle):
    name = 'buck/coul/long'
    pass


class BuckCoulLongCs(PairStyle):
    name = 'buck/coul/long/cs'
    pass


class BuckCoulMsm(PairStyle):
    name = 'buck/coul/msm'
    pass


class BuckLongCoulLong(PairStyle):
    name = 'buck/long/coul/long'
    pass


class Colloid(PairStyle):
    name = 'colloid'
    pass


class Comb(PairStyle):
    name = 'comb'
    pass


class Comb3(PairStyle):
    name = 'comb3'
    pass


class CoulCut(PairStyle):
    name = 'coul/cut'
    pass


class CoulDebye(PairStyle):
    name = 'coul/debye'
    pass


class CoulDsf(PairStyle):
    name = 'coul/dsf'
    pass


class CoulLong(PairStyle):
    name = 'coul/long'
    pass


class CoulLongCs(PairStyle):
    name = 'coul/long/cs'
    pass


class CoulMsm(PairStyle):
    name = 'coul/msm'
    pass


class CoulStreitz(PairStyle):
    name = 'coul/streitz'
    pass


class CoulWolf(PairStyle):
    name = 'coul/wolf'
    pass


class Dpd(PairStyle):
    name = 'dpd'
    pass


class DpdTstat(PairStyle):
    name = 'dpd/tstat'
    pass


class Dsmc(PairStyle):
    name = 'dsmc'
    pass


class Eam(PairStyle):
    name = 'eam'
    pass


class EamAlloy(PairStyle):
    name = 'eam/alloy'
    pass


class EamFs(PairStyle):
    name = 'eam/fs'
    pass


class Eim(PairStyle):
    name = 'eim'
    pass


class Gauss(PairStyle):
    name = 'gauss'
    pass


class Gayberne(PairStyle):
    name = 'gayberne'
    pass


class GranHertzHistory(PairStyle):
    name = 'gran/hertz/history'
    pass


class GranHooke(PairStyle):
    name = 'gran/hooke'
    pass


class GranHookeHistory(PairStyle):
    name = 'gran/hooke/history'
    pass


class HbondDreidingLj(PairStyle):
    name = 'hbond/dreiding/lj'
    pass


class HbondDreidingMorse(PairStyle):
    name = 'hbond/dreiding/morse'
    pass


class Kim(PairStyle):
    name = 'kim'
    pass


class Lcbop(PairStyle):
    name = 'lcbop'
    pass


class LineLj(PairStyle):
    name = 'line/lj'
    pass


class LjCharmmCoulCharmm(PairStyle):
    name = 'lj/charmm/coul/charmm'
    pass


class LjCharmmCoulCharmmImplicit(PairStyle):
    name = 'lj/charmm/coul/charmm/implicit'
    pass


class LjCharmmCoulLong(PairStyle):
    name = 'lj/charmm/coul/long'
    pass


class LjCharmmCoulMsm(PairStyle):
    name = 'lj/charmm/coul/msm'
    pass


class LjClass2(PairStyle):
    name = 'lj/class2'
    pass


class LjClass2CoulCut(PairStyle):
    name = 'lj/class2/coul/cut'
    pass


class LjClass2CoulLong(PairStyle):
    name = 'lj/class2/coul/long'
    pass


class LjCubic(PairStyle):
    name = 'lj/cubic'
    pass


class LjCut(PairStyle):
    name = 'lj/cut'
    pass


class LjCutCoulCut(PairStyle):
    name = 'lj/cut/coul/cut'
    pass


class LjCutCoulDebye(PairStyle):
    name = 'lj/cut/coul/debye'
    pass


class LjCutCoulDsf(PairStyle):
    name = 'lj/cut/coul/dsf'
    pass


class LjCutCoulLong(PairStyle):
    name = 'lj/cut/coul/long'
    pass


class LjCutCoulLongCs(PairStyle):
    name = 'lj/cut/coul/long/cs'
    pass


class LjCutCoulMsm(PairStyle):
    name = 'lj/cut/coul/msm'
    pass


class LjCutDipoleCut(PairStyle):
    name = 'lj/cut/dipole/cut'
    pass


class LjCutDipoleLong(PairStyle):
    name = 'lj/cut/dipole/long'
    pass


class LjCutTip4pCut(PairStyle):
    name = 'lj/cut/tip4p/cut'
    pass


class LjCutTip4pLong(PairStyle):
    name = 'lj/cut/tip4p/long'
    pass


class LjExpand(PairStyle):
    name = 'lj/expand'
    pass


class LjGromacs(PairStyle):
    name = 'lj/gromacs'
    pass


class LjGromacsCoulGromacs(PairStyle):
    name = 'lj/gromacs/coul/gromacs'
    pass


class LjLongCoulLong(PairStyle):
    name = 'lj/long/coul/long'
    pass


class LjLongDipoleLong(PairStyle):
    name = 'lj/long/dipole/long'
    pass


class LjLongTip4pLong(PairStyle):
    name = 'lj/long/tip4p/long'
    pass


class LjSmooth(PairStyle):
    name = 'lj/smooth'
    pass


class LjSmoothLinear(PairStyle):
    name = 'lj/smooth/linear'
    pass


class Lj96Cut(PairStyle):
    name = 'lj96/cut'
    pass


class Lubricate(PairStyle):
    name = 'lubricate'
    pass


class LubricatePoly(PairStyle):
    name = 'lubricate/poly'
    pass


class LubricateU(PairStyle):
    name = 'lubricate/u'
    pass


class LubricateUPoly(PairStyle):
    name = 'lubricate/u/poly'
    pass


class Meam(PairStyle):
    name = 'meam'
    pass


class MieCut(PairStyle):
    name = 'mie/cut'
    pass


class Morse(PairStyle):
    name = 'morse'
    pass


class Nb3bHarmonic(PairStyle):
    name = 'nb3b/harmonic'
    pass


class NmCut(PairStyle):
    name = 'nm/cut'
    pass


class NmCutCoulCut(PairStyle):
    name = 'nm/cut/coul/cut'
    pass


class NmCutCoulLong(PairStyle):
    name = 'nm/cut/coul/long'
    pass


class PeriEps(PairStyle):
    name = 'peri/eps'
    pass


class PeriLps(PairStyle):
    name = 'peri/lps'
    pass


class PeriPmb(PairStyle):
    name = 'peri/pmb'
    pass


class PeriVes(PairStyle):
    name = 'peri/ves'
    pass


class Polymorphic(PairStyle):
    name = 'polymorphic'
    pass


class Reax(PairStyle):
    name = 'reax'
    pass


class Rebo(PairStyle):
    name = 'rebo'
    pass


class Resquared(PairStyle):
    name = 'resquared'
    pass


class Snap(PairStyle):
    name = 'snap'
    pass


class Soft(PairStyle):
    name = 'soft'
    pass


class Sw(PairStyle):
    name = 'sw'
    pass


class Table(PairStyle):
    name = 'table'
    pass


class Tersoff(PairStyle):
    name = 'tersoff'
    pass


class TersoffMod(PairStyle):
    name = 'tersoff/mod'
    pass


class TersoffZbl(PairStyle):
    name = 'tersoff/zbl'
    pass


class Tip4pCut(PairStyle):
    name = 'tip4p/cut'
    pass


class Tip4pLong(PairStyle):
    name = 'tip4p/long'
    pass


class TriLj(PairStyle):
    name = 'tri/lj'
    pass


class Vashishta(PairStyle):
    name = 'vashishta'
    pass


class Yukawa(PairStyle):
    name = 'yukawa'
    pass


class YukawaColloid(PairStyle):
    name = 'yukawa/colloid'
    pass


class Zbl(PairStyle):
    name = 'zbl'
    pass




import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from abinitparser.metainfo import abinit_autogenerated
from nomad.datamodel.metainfo import public

m_package = Package(
    name='abinit_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='abinit.nomadmetainfo.json'))


class x_abinit_section_stress_tensor(MSection):
    '''
    Section describing the stress tensor
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_abinit_section_stress_tensor'))

    x_abinit_stress_tensor_xx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        xx component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_xx'))

    x_abinit_stress_tensor_yy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        yy component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_yy'))

    x_abinit_stress_tensor_zz = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        zz component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_zz'))

    x_abinit_stress_tensor_zy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        zy component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_zy'))

    x_abinit_stress_tensor_zx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        zx component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_zx'))

    x_abinit_stress_tensor_yx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        yx component of the stress tensor
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_stress_tensor_yx'))


class x_abinit_section_dataset_header(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_abinit_section_dataset_header'))

    x_abinit_dataset_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Dataset number
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_dataset_number'))

    x_abinit_vprim_1 = Quantity(
        type=str,
        shape=[],
        description='''
        Primitive axis 1
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vprim_1'))

    x_abinit_vprim_2 = Quantity(
        type=str,
        shape=[],
        description='''
        Primitive axis 2
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vprim_2'))

    x_abinit_vprim_3 = Quantity(
        type=str,
        shape=[],
        description='''
        Primitive axis 3
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vprim_3'))


class x_abinit_section_var(MSection):
    '''
    -
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_abinit_section_var'))

    x_abinit_vardtset = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Variable dataset number
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vardtset'))

    x_abinit_varname = Quantity(
        type=str,
        shape=[],
        description='''
        Variable name
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_varname'))

    x_abinit_varvalue = Quantity(
        type=str,
        shape=[],
        description='''
        Variable value
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_varvalue'))

    x_abinit_vartruncation = Quantity(
        type=str,
        shape=[],
        description='''
        Variable truncation length
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_vartruncation'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_abinit_parallel_compilation = Quantity(
        type=str,
        shape=[],
        description='''
        Parallel or sequential compilation
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_parallel_compilation'))

    x_abinit_start_date = Quantity(
        type=str,
        shape=[],
        description='''
        Start date as string
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_start_date'))

    x_abinit_start_time = Quantity(
        type=str,
        shape=[],
        description='''
        Start time as string
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_start_time'))

    x_abinit_input_file = Quantity(
        type=str,
        shape=[],
        description='''
        Input file name
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_input_file'))

    x_abinit_output_file = Quantity(
        type=str,
        shape=[],
        description='''
        Output file name
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_output_file'))

    x_abinit_input_files_root = Quantity(
        type=str,
        shape=[],
        description='''
        Root for input files
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_input_files_root'))

    x_abinit_output_files_root = Quantity(
        type=str,
        shape=[],
        description='''
        Root for output files
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_output_files_root'))

    x_abinit_total_cpu_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total CPU time
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_total_cpu_time'))

    x_abinit_total_wallclock_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total wallclock time
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_total_wallclock_time'))

    x_abinit_completed = Quantity(
        type=str,
        shape=[],
        description='''
        Message that the calculation was completed
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_completed'))

    x_abinit_section_var = SubSection(
        sub_section=SectionProxy('x_abinit_section_var'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_abinit_section_var'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_abinit_atom_xcart_final = Quantity(
        type=str,
        shape=[],
        description='''
        Cartesian coordinates of an atom at the end of the dataset
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_xcart_final'))

    x_abinit_atom_xcart = Quantity(
        type=str,
        shape=[],
        description='''
        Cartesian coordinates of an atom at the end of a single configuration calculation
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_xcart'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_abinit_magnetisation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total magnetisation.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_magnetisation'))

    x_abinit_fermi_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Fermi energy.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_fermi_energy'))

    x_abinit_single_configuration_calculation_converged = Quantity(
        type=str,
        shape=[],
        description='''
        Determines whether a single configuration calculation is converged.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_single_configuration_calculation_converged'))

    x_abinit_atom_force = Quantity(
        type=str,
        shape=[],
        description='''
        Force acting on an atom at the end of a single configuration calculation
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_force'))

    x_abinit_atom_force_final = Quantity(
        type=str,
        shape=[],
        description='''
        Force acting on an atom at the end of the dataset
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_atom_force_final'))

    x_abinit_energy_kinetic = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Kinetic energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_kinetic'))

    x_abinit_energy_hartree = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Hartree energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_hartree'))

    x_abinit_energy_xc = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        XC energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_xc'))

    x_abinit_energy_ewald = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Ewald energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_ewald'))

    x_abinit_energy_psp_core = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Pseudopotential core energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_psp_core'))

    x_abinit_energy_psp_local = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Local pseudopotential energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_psp_local'))

    x_abinit_energy_psp_nonlocal = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Non-local pseudopotential energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_psp_nonlocal'))

    x_abinit_energy_internal = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Internal energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_internal'))

    x_abinit_energy_ktentropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -kT*entropy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_ktentropy'))

    x_abinit_energy_band = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Band energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_band'))

    x_abinit_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total energy
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_energy_total'))

    x_abinit_section_stress_tensor = SubSection(
        sub_section=SectionProxy('x_abinit_section_stress_tensor'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_abinit_section_stress_tensor'))


class section_eigenvalues(public.section_eigenvalues):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_eigenvalues'))

    x_abinit_kpt = Quantity(
        type=str,
        shape=[],
        description='''
        The reduced coordinates of a k-point
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_kpt'))

    x_abinit_wtk = Quantity(
        type=str,
        shape=[],
        description='''
        A k-point weight
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_wtk'))

    x_abinit_eigenvalues = Quantity(
        type=str,
        shape=[],
        description='''
        List of eigenvalues on a given k-point
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_eigenvalues'))

    x_abinit_occupations = Quantity(
        type=str,
        shape=[],
        description='''
        List of occupations on a given k-point
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_occupations'))


class x_abinit_section_dataset(abinit_autogenerated.x_abinit_section_dataset):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='x_abinit_section_dataset'))

    x_abinit_geometry_optimization_converged = Quantity(
        type=str,
        shape=[],
        description='''
        Determines whether a geometry optimization is converged.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_geometry_optimization_converged'))

    x_abinit_eig_filename = Quantity(
        type=str,
        shape=[],
        description='''
        Name of file where the eigenvalues were written to.
        ''',
        a_legacy=LegacyDefinition(name='x_abinit_eig_filename'))

    x_abinit_section_dataset_header = SubSection(
        sub_section=SectionProxy('x_abinit_section_dataset_header'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_abinit_section_dataset_header'))


m_package.__init_metainfo__()
